﻿
using System;
using System.ComponentModel;
using System.IO;
using Styx.Helpers;
using Styx.WoWInternals.WoWObjects;

using DefaultValue = Styx.Helpers.DefaultValueAttribute;

namespace Singular.Settings
{
    public enum DeathKnightPresence
    {
        None = 0,
        Auto,
        Blood,
        Frost,
        Unholy
    }

    internal class DeathKnightSettings : Styx.Helpers.Settings
    {
        public DeathKnightSettings()
            : base(Path.Combine(SingularSettings.SingularSettingsPath, "DeathKnight.xml"))
        {
        }

        #region Common

        [Setting]
        [DefaultValue(DeathKnightPresence.Auto)]
        [Category("一般")]
        [DisplayName("灵气")]
        [Description("Auto: 选择当前专精/角色/状态下最合适的灵气, None:玩家选择")]
        public DeathKnightPresence Presence { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("一般")]
        [DisplayName("亡者大军")]
        public bool UseArmyOfTheDead { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("一般")]
        [DisplayName("冰霜之路")]
        public bool UsePathOfFrost { get; set; }

        [Setting]
        [DefaultValue(60)]
        [Category("一般")]
        [DisplayName("符能转换，生命 %")]
        [Description("使用符能转化治疗，生命 %")]
        public int ConversionPercent { get; set; }

        [Setting]
        [DefaultValue(80)]
        [Category("一般")]
        [DisplayName("符能转换，符文能量 %")]
        [Description("当符文能量高于设定的百分比时，才使用符文转换")]
        public int MinimumConversionRunicPowerPrecent { get; set; }

        [Setting]
        [DefaultValue(2)]
        [Category("一般")]
        [DisplayName("枯萎凋零数量")]
        [Description("当敌人数量达到设定值时，使用枯萎凋零")]
        public int DeathAndDecayCount { get; set; }

        [Setting]
        [DefaultValue(60)]
        [Category("一般")]
        [DisplayName("天灾契约，生命 %")]
        [Description("使用天灾契约治疗，生命 %")]
        public int DeathPactPercent { get; set; }

        [Setting]
        [DefaultValue(60)]
        [Category("一般")]
        [DisplayName("死亡虹吸，生命 %")]
        [Description("使用死亡虹吸治疗，生命 %")]
        public int DeathSiphonPercent { get; set; }

        [Setting]
        [DefaultValue(30)]
        [Category("一般")]
        [DisplayName("紧急使用灵界打击，生命 %[DPS天赋]")]
        public int DeathStrikeEmergencyPercent { get; set; }

        [Setting]
        [DefaultValue(30)]
        [Category("一般")]
        [DisplayName("冰封之韧，生命 %")]
        public int IceboundFortitudePercent { get; set; }

        [Setting]
        [DefaultValue(60)]
        [Category("一般")]
        [DisplayName("巫妖之躯，生命 %")]
        [Description("使用巫妖之躯+调零缠绕治疗，生命 %")]
        public int LichbornePercent { get; set; }


        #endregion

        #region Category: Blood

        [Setting]
        [DefaultValue(60)]
        [Category("鲜血")]
        [DisplayName("亡者大军，生命 %")]
        [Description("低于设置的生命百分比时使用亡者大军")]
        public int ArmyOfTheDeadPercent { get; set; }

        [Setting]
        [DefaultValue(4)]
        [Category("鲜血")]
        [DisplayName("血液沸腾数量")]
        [Description("使用血液沸腾，当附近的敌人数量达到。")]
        public int BloodBoilCount { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("鲜血")]
        [DisplayName("优先使用白骨之盾")]
        [Description("False:冷却时使用, True: 仅当当前没有白骨之盾，吸血鬼之血，符文武器刃舞，巫妖之躯，冰封之韧等效果存在时使用")]
        public bool BoneShieldExclusive { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("鲜血")]
        [DisplayName("优先使用天灾契约")]
        [Description("False:需要时使用召唤食尸鬼来使用天灾契约, True: 召唤食尸鬼来使用天灾契约仅当当前没有白骨之盾，吸血鬼之血，符文武器刃舞，巫妖之躯，冰封之韧等效果存在时")]
        public bool DeathPactExclusive { get; set; }

        [Setting]
        [DefaultValue(60)]
        [Category("鲜血")]
        [DisplayName("符文武器增效，生命 %")]
        [Description("低于设置的生命百分比时使用符文武器增效")]
        public int EmpowerRuneWeaponPercent { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("鲜血")]
        [DisplayName("优先使用冰封之韧")]
        [Description("False:需要时使用, True:在需要时使用并且当前没有白骨之盾，吸血鬼之血，符文武器刃舞，巫妖之躯，冰封之韧等效果存在")]
        public bool IceboundFortitudeExclusive { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("鲜血")]
        [DisplayName("优先使用巫妖之躯")]
        [Description("仅当没有可用的白骨之盾、血族之裔、符文武器刃舞、冰封之韧时才使用巫妖之躯")]
        public bool LichborneExclusive { get; set; }

        [Setting]
        [DefaultValue(60)]
        [Category("鲜血")]
        [DisplayName("符文分流，生命 %")]
        [Description("使用符文分流治疗，生命 %")]
        public int RuneTapPercent { get; set; }

        [Setting]
        [DefaultValue(70)]
        [Category("鲜血")]
        [DisplayName("召唤食尸鬼，生命 %")]
        [Description("鲜血专精下，低于设置的生命百分比时使用召唤食尸鬼")]
        public int SummonGhoulPercentBlood { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("鲜血")]
        [DisplayName("把召唤食尸鬼作为输出技能")]
        [Description("把召唤食尸鬼作为输出技能")]
        public bool UseGhoulAsDpsCdBlood { get; set; }

        [Setting]
        [DefaultValue(60)]
        [Category("鲜血")]
        [DisplayName("吸血鬼之血，生命 %")]
        [Description("低于设置的生命百分比时使用吸血鬼之血")]
        public int VampiricBloodPercent { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("鲜血")]
        [DisplayName("优先使用吸血鬼之血")]
        [Description("False: 在需要时使用, True: 在需要时使用并且当前没有白骨之盾，吸血鬼之血，符文武器刃舞，巫妖之躯，冰封之韧等效果存在")]
        public bool VampiricBloodExclusive { get; set; } 
        #endregion

        #region Category: Frost
        [Setting]
        [DefaultValue(false)]
        [Category("冰霜")]
        [DisplayName("把召唤食尸鬼作为输出技能")]
        [Description("把召唤食尸鬼作为输出技能")]
        public bool UseGhoulAsDpsCdFrost { get; set; }

        [Setting]
        [DefaultValue(70)]
        [Category("冰霜")]
        [DisplayName("召唤食尸鬼，生命 %")]
        [Description("冰霜专精下，低于设置的生命百分比时使用召唤食尸鬼")]
        public int SummonGhoulPercentFrost { get; set; }
        
        #endregion

        #region Category: Unholy

        [Setting]
        [DefaultValue(true)]
        [Category("邪恶")]
        [DisplayName("召唤石像鬼")]
        [Description("False: 不使用该技能, True: 当需要使用长CD技能时使用 (首领战, PVP战, 有压力的单人战斗)")]
        public bool UseSummonGargoyle { get; set; }

        #endregion

    }
}