﻿
using System.ComponentModel;
using System.IO;
using Styx.Helpers;
using Styx.WoWInternals.WoWObjects;

using DefaultValue = Styx.Helpers.DefaultValueAttribute;
using System.Windows.Forms;

namespace Singular.Settings
{
    internal class HotkeySettings : Styx.Helpers.Settings
    {
        public HotkeySettings()
            : base(Path.Combine(SingularSettings.CharacterSettingsPath, "SingularSettings.Hotkeys.xml"))
        { 
            // bit of a hack -- SavedToFile setting tracks if we have ever saved
            // .. these settings.  this is needed because we can't use the DefaultValue
            // .. attribute for a multi values setting
            if (!SavedToFile)
            {
                // force to true so when loaded next time we skip the following
                SavedToFile = true;

                // set default values for array
                SuspendMovementKeys = new Keys[] 
                {
                    // default WOW bindings for movement keys
                    Keys.MButton,
                    Keys.RButton,
                    Keys.W,
                    Keys.S,
                    Keys.A,
                    Keys.D,
                    Keys.Q,
                    Keys.E,
                    Keys.Up,
                    Keys.Down,
                    Keys.Left,
                    Keys.Right
                };
            }
        }


        // hidden setting to track if we have ever saved this settings file before
        [Setting]
        [Browsable(false)]
        [DefaultValue(false)]
        public bool SavedToFile { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("行为信息")]
        [DisplayName("聊天框信息")]
        [Description("当热键按下时在聊天框显示信息")]
        public bool ChatFrameMessage { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("控制")]
        [DisplayName("暂停热键行为")]
        [Description("True: 按下禁用，再按启用; False:按住禁用，松开启用")]
        public bool KeysToggleBehavior { get; set; }

        [Setting]
        [DefaultValue(Keys.None)]
        [Category("行为热键")]
        [DisplayName("AOE技能热键")]
        [Description("允许/禁用AOE战斗技能")]
        public Keys AoeToggle { get; set; }

        [Setting]
        [DefaultValue(Keys.None)]
        [Category("行为热键")]
        [DisplayName("战斗热键")]
        [Description("允许/禁用所有战斗技能")]
        public Keys CombatToggle { get; set; }

        [Setting]
        [DefaultValue(Keys.None)]
        [Category("行为热键")]
        [DisplayName("移动热键")]
        [Description("允许/禁用Singular控制移动")]
        public Keys MovementToggle { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("临时行为")]
        [DisplayName("暂停移动")]
        [Description("True:下面设置的移动控制热键可以暂停软件控制角色移动 # 秒")]
        public bool SuspendMovement { get; set; }

        [Setting]
        [DefaultValue(3)]
        [Category("临时行为")]
        [DisplayName("暂停时间")]
        [Description("设置暂停热键的暂停时间<秒>，从松开按键的那一刻开始算起")]
        public int SuspendDuration { get; set; }

        [Setting]
        [Category("临时行为")]
        [DisplayName("暂停热键")]
        [Description("按键会暂停所有的移动控制 # 秒")]
        public Keys[] SuspendMovementKeys { get; set; }

    }
}