﻿
using System.ComponentModel;
using System.IO;
using Singular.ClassSpecific.Rogue;
using Styx.Helpers;
using DefaultValue = Styx.Helpers.DefaultValueAttribute;

namespace Singular.Settings
{
    internal class MonkSettings : Styx.Helpers.Settings
    {
        public MonkSettings()
            : base(Path.Combine(SingularSettings.SingularSettingsPath, "Monk.xml")) { }

        #region Spheres

        [Setting]
        [DefaultValue(true)]
        [Category("珠子")]
        [DisplayName("移动到珠子")]
        [Description("允许移动到珠子附近治疗或者回气")]
        public bool MoveToSpheres { get; set; }

        [Setting]
        [DefaultValue(15)]
        [Category("珠子")]
        [DisplayName("战斗外捡珠子的最大距离")]
        [Description("不在战斗状态下捡珠子的最大距离")]
        public int SphereDistanceAtRest { get; set; }

        [Setting]
        [DefaultValue(5)]
        [Category("珠子")]
        [DisplayName("战斗中捡珠子的最大距离")]
        [Description("设置战斗中捡珠子的最大距离")]
        public int SphereDistanceInCombat { get; set; }

        [Setting]
        [DefaultValue(45)]
        [Category("珠子")]
        [DisplayName("战斗外使用疗伤珠")]
        [Description("不在战斗状态下低于设置的生命百分比时使用疗伤珠治疗")]
        public int RestHealingSphereHealth { get; set; }

        #endregion

        #region Common

        [Setting]
        [DefaultValue(60)]
        [Category("一般")]
        [DisplayName("壮胆酒，生命 %")]
        [Description("低于设置的生命百分比时使用壮胆酒")]
        public float FortifyingBrewPercent { get; set; }

        [Setting]
        [DefaultValue(70)]
        [Category("一般")]
        [DisplayName("真气波，生命 %")]
        [Description("低于设置的生命百分比时使用真气波")]
        public float ChiWavePercent { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("一般")]
        [DisplayName("使用探云手")]
        [Description("True: 冷却结束就使用探云手; False:不使用该技能")]
        public bool UseGrappleWeapon { get; set; }


        #endregion

        #region Brewmaster

        [Setting]
        [DefaultValue(70)]
        [Category("酒仙")]
        [DisplayName("慈悲庇护（团队血量）")]
        [Description("当团队血量百分比低于设置的数字时使用慈悲庇护")]
        public int AvertHarmGroupHealthPct { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("酒仙")]
        [DisplayName("使用慈悲庇护")]
        public bool UseAvertHarm { get; set; }

        [Setting]
        [DefaultValue(6)]
        [Category("酒仙")]
        [DisplayName("使用飘渺酒的最小层数")]
        [Description("当角色获得飘渺酒的层数达到或高于设置的数字时使用飘渺酒")]
        public int ElusiveBrewMinumumCount { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("酒仙")]
        [DisplayName("使用飘渺酒")]
        public bool UseElusiveBrew { get; set; }
        #endregion


        [Setting]
        [DefaultValue(true)]
        [Category("踏风")]
        [DisplayName("允许协助治疗")]
        public bool AllowOffHeal { get; set; }

        [Setting]
        [DefaultValue(65)]
        [Category("踏风")]
        [DisplayName("移花接木治疗")]
        public int ExpelHarmHealth { get; set; }
    }
}