﻿
using System;
using System.ComponentModel;
using System.IO;
using Singular.ClassSpecific.Paladin;

using Styx.Helpers;

using DefaultValue = Styx.Helpers.DefaultValueAttribute;
using Styx.CommonBot;
using Styx;

namespace Singular.Settings
{

    public enum PaladinBlessings
    {
        Auto,
        Kings,
        Might
    }

    public enum PaladinSeal
    {
        None = 0,
        Auto = 1,
        Command,
        Truth,
        Insight,
        Righteousness,
        Justice
    }

    internal class PaladinSettings : Styx.Helpers.Settings
    {
        public PaladinSettings()
            : base(Path.Combine(SingularSettings.SingularSettingsPath, "Paladin.xml"))
        {
        }


        #region Common
        /*
        [Setting]
        [DefaultValue(PaladinAura.Auto)]
        [Category("Common")]
        [DisplayName("Aura")]
        [Description("The aura to be used while not mounted. Set this to Auto to allow the CC to automatically pick the aura depending on spec.")]
        public PaladinAura Aura { get; set; }
        */

        [Setting]
        [DefaultValue(90)]
        [Category("一般")]
        [DisplayName("圣光术，治疗 %")]
        [Description("低于设置的生命百分比时使用圣光术")]
        public int HolyLightHealth { get; set; }

        [Setting]
        [DefaultValue(PaladinBlessings.Auto)]
        [Category("一般")]
        [DisplayName("祝福")]
        [Description("使用何种祝福,力量/王者 (Auto: 最佳选择)")]
        public PaladinBlessings Blessings { get; set; }

        [Setting]
        [DefaultValue(PaladinSeal.Auto)]
        [Category("一般")]
        [DisplayName("圣印")]
        [Description("使用何种圣印 (None: 玩家控制, Auto: 最佳选择)")]
        public PaladinSeal Seal { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("一般")]
        [DisplayName("制裁之锤")]
        [Description("在普通状态冷却时间一到就使用制裁之锤以减少受到的伤害")]
        public bool StunMobsWhileSolo { get; set; }

        #endregion

        #region Self-Heal

        [Setting]
        [DefaultValue(25)]
        [Category("自疗")]
        [DisplayName("圣疗，生命 %")]
        [Description("低于设置的生命百分比时使用圣疗")]
        public int SelfLayOnHandsHealth { get; set; }

        [Setting]
        [DefaultValue(40)]
        [Category("自疗")]
        [DisplayName("圣光闪现，治疗 %")]
        [Description("低于设置的生命百分比时使用圣光闪现")]
        public int SelfFlashOfLightHealth { get; set; }

        [Setting]
        [DefaultValue(0)]
        [Category("自疗")]
        [DisplayName("1豆荣耀圣令")]
        [Description("低于设置的生命百分比时使用荣耀圣令如果仅有1格圣能时")]
        public int SelfWordOfGloryHealth1 { get; set; }

        [Setting]
        [DefaultValue(0)]
        [Category("自疗")]
        [DisplayName("2豆荣耀圣令")]
        [Description("低于设置的生命百分比时使用荣耀圣令当有2格圣能时")]
        public int SelfWordOfGloryHealth2 { get; set; }

        [Setting]
        [DefaultValue(65)]
        [Category("自疗")]
        [DisplayName("3豆荣耀圣令")]
        [Description("低于设置的生命百分比时使用荣耀圣令当有3+格圣能时")]
        public int SelfWordOfGloryHealth3 { get; set; }

        [Setting]
        [DefaultValue(65)]
        [Category("自疗")]
        [DisplayName("永恒之火，生命 %")]
        [Description("低于设置的生命百分比时使用永恒之火当有3+格圣能时")]
        public int SelfEternalFlameHealth { get; set; }

        #endregion  

        #region Holy

        [Setting]
        [DefaultValue(25)]
        [Category("神圣")]
        [DisplayName("圣疗，生命 %")]
        [Description("低于设置的生命百分比时使用圣疗")]
        public int LayOnHandsHealth { get; set; }

        [Setting]
        [DefaultValue(50)]
        [Category("神圣")]
        [DisplayName("圣光闪现，生命 %")]
        [Description("低于设置的生命百分比时使用圣光闪现")]
        public int FlashOfLightHealth { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("神圣")]
        [DisplayName("保持坦克的永恒之火")]
        [Description("维持坦克身上的永恒之火状态")]
        public bool KeepEternalFlameUp { get; set; } 

        [Setting]
        [DefaultValue(80)]
        [Category("神圣")]
        [DisplayName("黎明圣光，生命 %")]
        [Description("低于设置的生命百分比时使用黎明圣光")]
        public int LightOfDawnHealth { get; set; }

        [Setting]
        [DefaultValue(2)]
        [Category("神圣")]
        [DisplayName("黎明圣光，数量")]
        [Description("超过设置数量的队友低于设置的生命百分比时使用黎明圣光")]
        public int LightOfDawnCount { get; set; }

        [Setting]
        [DefaultValue(90)]
        [Category("神圣")]
        [DisplayName("神圣震击，生命 %")]
        [Description("低于设置的生命百分比时使用神圣震击")]
        public int HolyShockHealth { get; set; }

        [Setting]
        [DefaultValue(65)]
        [Category("神圣")]
        [DisplayName("神圣之光，生命 %")]
        [Description("低于设置的生命百分比时使用神圣之光")]
        public int DivineLightHealth { get; set; }

        [Setting]
        [DefaultValue(50)]
        [Category("神圣")]
        [DisplayName("神圣恳求，魔法 %")]
        [Description("低于设置的魔法百分比时使用神圣恳求")]
        public double DivinePleaMana { get; set; }

         #endregion

        #region Protection
        [Setting]
        [DefaultValue(40)]
        [Category("防护")]
        [DisplayName("远古列王守卫，生命 %")]
        [Description("低于设置的生命百分比时使用远古列王守卫")]
        public int GoAKHealth { get; set; }

        [Setting]
        [DefaultValue(40)]
        [Category("防护")]
        [DisplayName("炽热防御者，生命 %")]
        [Description("低于设置的生命百分比时使用炽热防御者")]
        public int ArdentDefenderHealth { get; set; }

        [Setting]
        [DefaultValue(80)]
        [Category("防护")]
        [DisplayName("圣佑术，生命 %")]
        [Description("低于设置的生命百分比时使用圣佑术")]
        public int DivineProtectionHealthProt { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("防护")]
        [DisplayName("复仇者之盾引怪")]
        [Description("仅仅使用复仇者之盾来引怪")]
        public bool AvengersPullOnly { get; set; }

        [Setting]
        [DefaultValue(3)]
        [Category("防护")]
        [DisplayName("奉献，数量")]
        [Description("超过设置的敌人数量攻击你时使用奉献")]
        public int ProtConsecrationCount { get; set; }
        #endregion

        #region Retribution
        [Setting]
        [DefaultValue(70)]
        [Category("惩戒")]
        [DisplayName("圣佑术，生命 %")]
        [Description("低于设置的生命百分比时使用圣佑术")]
        public int DivineProtectionHealthRet { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("惩戒")]
        [DisplayName("使用复仇之怒/神圣复仇者/远古列王守卫")]
        [Description("True:自动使用复仇之怒，神圣复仇者，远古列王守卫。False: 你自己来决定何时使用。")]
        public bool RetAvengAndGoatK { get; set; }

       #endregion
    }
}