﻿
using System;
using System.ComponentModel;
using System.IO;
using Styx.Helpers;

using DefaultValue = Styx.Helpers.DefaultValueAttribute;
using Styx;
using System.Drawing;

namespace Singular.Settings
{
    internal class PriestSettings : Styx.Helpers.Settings
    {
        public PriestSettings()
            : base(Path.Combine(SingularSettings.SingularSettingsPath, "Priest.xml"))
        {
        }

        #region Holy Healing Context Late Loading Wrappers

        private HolyPriestHealSettings _holybattleground;
        private HolyPriestHealSettings _holyinstance;
        private HolyPriestHealSettings _holyraid;

        [Browsable(false)]
        public HolyPriestHealSettings HolyBattleground { get { return _holybattleground ?? (_holybattleground = new HolyPriestHealSettings(HealingContext.Battlegrounds)); } }

        [Browsable(false)]
        public HolyPriestHealSettings HolyInstance { get { return _holyinstance ?? (_holyinstance = new HolyPriestHealSettings(HealingContext.Instances)); } }

        [Browsable(false)]
        public HolyPriestHealSettings HolyRaid { get { return _holyraid ?? (_holyraid = new HolyPriestHealSettings(HealingContext.Raids)); } }

        [Browsable(false)]
        public HolyPriestHealSettings HolyHeal { get { return HolyHealLookup(Singular.SingularRoutine.CurrentWoWContext); } }

        public HolyPriestHealSettings HolyHealLookup(WoWContext ctx)
        {
            if (ctx == WoWContext.Instances)
                return StyxWoW.Me.CurrentMap.IsRaid ? HolyRaid : HolyInstance;
            return HolyBattleground;
        }

        #endregion


        #region Disc Healing Context Late Loading Wrappers

        private DiscPriestHealSettings _discbattleground;
        private DiscPriestHealSettings _discinstance;
        private DiscPriestHealSettings _discraid;

        [Browsable(false)]
        public DiscPriestHealSettings DiscBattleground { get { return _discbattleground ?? (_discbattleground = new DiscPriestHealSettings(HealingContext.Battlegrounds)); } }

        [Browsable(false)]
        public DiscPriestHealSettings DiscInstance { get { return _discinstance ?? (_discinstance = new DiscPriestHealSettings(HealingContext.Instances)); } }

        [Browsable(false)]
        public DiscPriestHealSettings DiscRaid { get { return _discraid ?? (_discraid = new DiscPriestHealSettings(HealingContext.Raids)); } }

        [Browsable(false)]
        public DiscPriestHealSettings DiscHeal { get { return DiscHealLookup(Singular.SingularRoutine.CurrentWoWContext); } }

        public DiscPriestHealSettings DiscHealLookup(WoWContext ctx)
        {
            if (ctx == WoWContext.Instances)
                return StyxWoW.Me.CurrentMap.IsRaid ? DiscRaid : DiscInstance;
            return DiscBattleground;
        }

        #endregion

        #region Shadow

        [Setting]
        [DefaultValue(30)]
        [Category("暗影")]
        [DisplayName("快速治疗，生命 %")]
        [Description("低于设置的生命百分比时使用快速治疗（暗影天赋）")]
        public double ShadowFlashHealHealth { get; set; }

        [Setting]
        [DefaultValue(50)]
        [Category("暗影")]
        [DisplayName("精神鞭笞，魔法 %")]
        [Description("高于设置的魔法百分比时仅使用精神鞭笞")]
        public double MindFlayMana { get; set; }

        [Setting]
        [DefaultValue(60)]
        [Category("暗影")]
        [DisplayName("真言术：盾，生命 %")]
        [Description("低于设置的生命百分比时使用真言术：盾")]
        public double ShieldHealthPercent { get; set; }

/*
        [Setting]
        [DefaultValue(15)]
        [Category("Shadow")]
        [DisplayName("Mind Blast Timer")]
        [Description("Casts mind blast anyway after this many seconds if we haven't got 3 shadow orbs")]
        public int MindBlastTimer { get; set; }
*/
        [Setting]
        [DefaultValue(2)]
        [Category("暗影")]
        [DisplayName("单人，暗影宝珠")]
        [Description("在一般情况下使用噬灵疫病时暗影宝珠的数量")]
        public int NormalContextOrbs { get; set; }
/*
        [Setting]
        [DefaultValue(false)]
        [Category("Shadow")]
        [DisplayName("Devouring Plague First")]
        [Description("Casts devouring plague before anything else, useful for farming low hp mobs")]
        public bool DevouringPlagueFirst { get; set; }
/*
        [Setting]
        [DefaultValue(false)]
        [Category("Shadow")]
        [DisplayName("Archangel on 5")]
        [Description("Always archangel on 5 dark evangelism, ignoring mana %")]
        public bool AlwaysArchangel5 { get; set; }
*/

        [Setting]
        [DefaultValue(20)]
        [Category("暗影")]
        [DisplayName("消散，魔法 %")]
        [Description("低于设置的魔法百分比时使用消散")]
        public int DispersionMana { get; set; }

        [Setting]
        [DefaultValue(40)]
        [Category("暗影")]
        [DisplayName("治疗法术")]
        [Description("低于设置的生命百分比时使用治疗法术治疗")]
        public int DontHealPercent { get; set; }
/*
        [Setting]
        [DefaultValue(true)]
        [Category("Shadow")]
        [DisplayName("No Shadowform Below Heal")]
        [Description("Won't attempt to re-enter shadowform while below healing spells health threshold")]
        public bool DontShadowFormHealth { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("Shadow")]
        [DisplayName("Psychic Horror Adds")]
        [Description("Attempt to psychic horror adds")]
        public bool UsePsychicHorrorAdds { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("Shadow")]
        [DisplayName("Psychic Horror Interrupt")]
        [Description("Attempt to psychic horror target as interrupt (on top of silence)")]
        public bool UsePsychicHorrorInterrupt { get; set; }
*/
        #endregion

        #region Common

        [Setting]
        [DefaultValue(true)]
        [Category("一般")]
        [DisplayName("开怪前上盾")]
        [Description("开怪前使用真言术：盾。这个设置不会影响到AOE输出循环。")]
        public bool UseShieldPrePull { get; set; }

        [Setting]
        [DefaultValue(00)]
        [Category("治疗")]
        [DisplayName("群体驱散，数量")]
        [Description("使用群体驱散的最小数量 (0: 禁用群体驱散)")]
        public int CountMassDispel { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("一般")]
        [DisplayName("使用渐隐术")]
        [Description("True: 获得仇恨时使用渐隐术 -或者- 点了幻影天赋时移除移动限制")]
        public bool UseFade { get; set; }

        /*
                        [Setting]
                        [DefaultValue(true)]
                        [Category("Common")]
                        [DisplayName("Use Shadow Protection")]
                        [Description("Use Shadow Protection buff")]
                        public bool UseShadowProtection { get; set; }
                */
        [Setting]
        [DefaultValue(true)]
        [Category("一般")]
        [DisplayName("使用心灵尖啸")]
        [Description("使用心灵尖啸")]
        public bool UsePsychicScream { get; set; }

        [Setting]
        [DefaultValue(2)]
        [Category("一般")]
        [DisplayName("使用心灵尖啸的敌人数量")]
        [Description("超过设置的敌人数量时使用心灵尖啸(不包含当前目标)")]
        public int PsychicScreamAddCount { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("一般")]
        [DisplayName("使用防护恐惧结界")]
        [Description("保持防护恐惧结界状态")]
        public bool UseFearWard { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("一般")]
        [DisplayName("使用心灵之火")]
        [Description("使用心灵之火, 否则使用心灵意志")]
        public bool UseInnerFire { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("一般")]
        [DisplayName("允许<风筝>")]
        [Description("战斗中移动来减少受到的伤害")]
        public bool AllowKiting { get; set; }

/*
                [Setting]
                [DefaultValue(75)]
                [Category("Common")]
                [DisplayName("Archangel Mana")]
                [Description("Archangel will be used at this value")]
                public int ArchangelMana { get; set; }
        */

        [Setting]
        [DefaultValue(75)]
        [Category("一般")]
        [DisplayName("暗影魔，魔法 %")]
        [Description("低于设置的魔法百分比时使用暗影魔")]
        public int ShadowfiendMana { get; set; }
/*
        [Setting]
        [DefaultValue(50)]
        [Category("Common")]
        [DisplayName("Mindbender Mana")]
        [Description("Mindbender will be used at this value")] 
        public int MindbenderMana { get; set; }
*/
        [Setting]
        [DefaultValue(50)]
        [Category("一般")]
        [DisplayName("希望圣歌，魔法 %")]
        [Description("低于设置的魔法百分比时使用希望圣歌")]
        public int HymnofHopeMana { get; set; }

        [Setting]
        [DefaultValue(30)]
        [Category("一般")]
        [DisplayName("绝望祷言，生命 %")]
        [Description("低于设置的生命百分比时使用绝望祷言")]
        public int DesperatePrayerHealth { get; set; }
                       

        //[Setting]
        //[DefaultValue(false)]
        //[Category("Common")]
        //[DisplayName("Use Wand")]
        //[Description("Uses wand if we're oom")]
        //public bool UseWand { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("一般")]
        [DisplayName("使用加速技能")]
        [Description("试图脱离战斗时使用第二排天赋的技能")]
        public bool UseSpeedBuff { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("一般")]
        [DisplayName("对tank使用速度爆发")]
        [Description("对tank使用第二层移动速度增益buff，当他们不在战斗且移动时")]
        public bool UseSpeedBuffOnTank { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("一般")]
        [DisplayName("使用信仰之跃")]
        [Description("当队友快死于环境伤害或者离近战敌人过近时对其使用信仰之跃")]
        public bool UseLeapOfFaith { get; set; }

        #endregion

        #region Discipline
/*
        [Setting]
        [DefaultValue(80)]
        [Category("戒律")]
        [DisplayName("灵魂护壳，生命 %")]
        [Description("高于设置的生命百分比时通过使用治疗祷言为队友添加灵魂护壳")]
        public int SpiritShell { get; set; }

        [Setting]
        [DefaultValue(30)]
        [Category("戒律")]
        [DisplayName("苦修，生命 %")]
        [Description("低于设置的生命百分比时使用苦修")]
        public int Penance { get; set; }

        [Setting]
        [DefaultValue(40)]
        [Category("戒律")]
        [DisplayName("快速治疗，生命 %")]
        [Description("低于设置的生命百分比时使用快速治疗")]
        public int FlashHeal { get; set; }

        [Setting]
        [DefaultValue(50)]
        [Category("戒律")]
        [DisplayName("强效治疗术，生命 %")]
        [Description("低于设置的生命百分比时使用强效治疗术")]
        public int GreaterHeal { get; set; }

        [Setting]
        [DefaultValue(70)]
        [Category("戒律")]
        [DisplayName("治疗术，生命 %")]
        [Description("低于设置的生命百分比时使用治疗术")]
        public int HealHealthPercentage { get; set; }

        [Setting]
        [DefaultValue(70)]
        [Category("戒律")]
        [DisplayName("恢复，生命 %")]
        [Description("低于设置的生命百分比时使用恢复")]
        public int Renew { get; set; }

        [Setting]
        [DefaultValue(30)]
        [Category("戒律")]
        [DisplayName("痛苦压制，生命 %")]
        [Description("低于设置的生命百分比时使用痛苦压制")]
        public int PainSuppression { get; set; }

        [Setting]
        [DefaultValue(70)]
        [Category("戒律")]
        [DisplayName("联结治疗，自己生命 %")]
        [Description("低于设置的自己的生命百分比时对自己使用联结治疗")]
        public int BindingHealMe { get; set; }

        [Setting]
        [DefaultValue(70)]
        [Category("戒律")]
        [DisplayName("联结治疗，队友生命 %")]
        [Description("低于设置的队友的生命百分比时对队友使用联结治疗")]
        public int BindingHealThem { get; set; }

        [Setting]
        [DefaultValue(50)]
        [Category("戒律")]
        [DisplayName("治疗祷言，生命 %")]
        [Description("低于设置的生命百分比时使用治疗祷言")]
        public int PrayerOfHealing { get; set; }

        [Setting]
        [DefaultValue(40)]
        [Category("戒律")]
        [DisplayName("真言术：盾，生命 %")]
        [Description("低于设置的生命百分比时使用真言术：盾")]
        public int PowerWordShield { get; set; }

        [Setting]
        [DefaultValue(3)]
        [Category("戒律")]
        [DisplayName("治疗祷言，数量")]
        [Description("超过设置数量的队友低于设置的生命百分比时使用治疗祷言")]
        public int PrayerOfHealingCount { get; set; }

        [Setting]
        [DefaultValue(70)]
        [Category("戒律")]
        [DisplayName("输出，魔法 %")]
        [Description("高于设置的魔法百分比尽可能的多输出(在队伍中使用)")]
        public int DpsMana { get; set; }
*/
        #endregion

        #region Holy
/*
        [Setting]
        [DefaultValue(95)]
        [Category("神圣")]
        [DisplayName("治疗术，生命 %")]
        [Description("低于设置的生命百分比时使用治疗术")]
        public int HolyHealPercent { get; set; }

        [Setting]
        [DefaultValue(50)]
        [Category("神圣")]
        [DisplayName("强效治疗术，生命 %")]
        [Description("低于设置的生命百分比时使用强效治疗术")]
        public int HolyGreaterHeal { get; set; }

        [Setting]
        [DefaultValue(25)]
        [Category("神圣")]
        [DisplayName("快速治疗，生命 %")]
        [Description("低于设置的生命百分比时使用快速治疗")]
        public int HolyFlashHeal { get; set; }

        [Setting]
        [DefaultValue(40)]
        [Category("神圣")]
        [DisplayName("神圣赞美诗，生命 %")]
        [Description("低于设置的生命百分比时使用神圣赞美诗")]
        public int DivineHymnHealth { get; set; }

        [Setting]
        [DefaultValue(6)]
        [Category("神圣")]
        [DisplayName("神圣赞美诗，数量")]
        [Description("超过设置数量的队友低于设置的生命百分比时使用神圣赞美诗")]
        public int DivineHymnCount { get; set; }

        [Setting]
        [DefaultValue(80)]
        [Category("神圣")]
        [DisplayName("治疗祷言（妙手回春），生命 %")]
        [Description("低于设置的生命百分比时使用治疗祷言（妙手回春）")]
        public int PrayerOfHealingSerendipityHealth { get; set; }

        [Setting]
        [DefaultValue(4)]
        [Category("神圣")]
        [DisplayName("治疗祷言（妙手回春），数量")]
        [Description("超过设置数量的队友低于设置的生命百分比时使用治疗祷言（妙手回春）")]
        public int PrayerOfHealingSerendipityCount { get; set; }

        [Setting]
        [DefaultValue(95)]
        [Category("神圣")]
        [DisplayName("治疗之环，生命 %")]
        [Description("低于设置的生命百分比时使用治疗之环")]
        public int CircleOfHealingHealth { get; set; }

        [Setting]
        [DefaultValue(5)]
        [Category("神圣")]
        [DisplayName("治疗之环，数量")]
        [Description("超过设置数量的队友低于设置的生命百分比时使用治疗之环")]
        public int CircleOfHealingCount { get; set; }

*/


        #endregion
    }

    internal class HolyPriestHealSettings : Singular.Settings.HealerSettings
    {
        private HolyPriestHealSettings()
            : base("", HealingContext.None)
        {
        }

        public HolyPriestHealSettings(HealingContext ctx)
            : base("Priest-Holy", ctx)
        {

            // we haven't created a settings file yet,
            //  ..  so initialize values for various heal contexts

            if (!SavedToFile)
            {
                if (ctx == Singular.HealingContext.Battlegrounds)
                {
                    Renew = 95;
                    PrayerOfMending = 94;
                    Heal = 0;
                    GreaterHeal = 0;
                    FlashHeal = 80;
                    BindingHeal = 55;
                    HolyLevel90Talent = 85;
                    HolyWordSanctuary = 95;
                    HolyWordSerenity = 85;
                    CircleOfHealing = 90;
                    PrayerOfHealing = 85;
                    DivineHymn = 75;
                    GuardianSpirit = 35;
                    VoidShift = 15;
                    CountHolyWordSanctuary = 3;
                    CountLevel90Talent = 3;
                    CountCircleOfHealing = 3;
                    CountPrayerOfHealing = 3;
                    CountDivineHymn = 4;
                }
                else if (ctx == Singular.HealingContext.Instances)
                {
                    Renew = 95;
                    PrayerOfMending = 85;
                    Heal = 90;
                    GreaterHeal = 65;
                    FlashHeal = 65;
                    BindingHeal = 35;
                    HolyLevel90Talent = 80;
                    HolyWordSanctuary = 0;
                    HolyWordSerenity = 85;
                    CircleOfHealing = 93;
                    PrayerOfHealing = 85;
                    DivineHymn = 75;
                    GuardianSpirit = 20;
                    VoidShift = 19;
                    CountHolyWordSanctuary = 3;
                    CountLevel90Talent = 3;
                    CountCircleOfHealing = 3;
                    CountPrayerOfHealing = 3;
                    CountDivineHymn = 4;
                }
                else if (ctx == Singular.HealingContext.Raids)
                {
                    PowerWordShield = 100;
                    Renew = 95;
                    PrayerOfMending = 94;
                    Heal = 0;
                    GreaterHeal = 40;
                    FlashHeal = 20;
                    BindingHeal = 30;
                    HolyLevel90Talent = 80;
                    HolyWordSanctuary = 92;
                    HolyWordSerenity = 0;
                    CircleOfHealing = 93;
                    PrayerOfHealing = 85;
                    DivineHymn = 75;
                    GuardianSpirit = 20;
                    VoidShift = 19;
                    CountHolyWordSanctuary = 3;
                    CountLevel90Talent = 4;
                    CountCircleOfHealing = 4;
                    CountPrayerOfHealing = 3;
                    CountDivineHymn = 5;
                }
                // omit case for WoWContext.Normal and let it use DefaultValue() values
            }

            // adjust Flash Heal if we have not previously 
            if (!FlashHealAdjusted && StyxWoW.Me.Level >= 34 && (ctx == HealingContext.Instances || ctx == HealingContext.Raids))
            {
                if (SavedToFile)
                    Logger.Write(Color.White, "Flash Heal % changed from {0} to {1} for {2}.  Visit Class Config and Save to make permanent.", FlashHeal, GuardianSpirit + 1, ctx.ToString());

                FlashHeal = GuardianSpirit + 1;
                FlashHealAdjusted = true;
            }

            SavedToFile = true;
        }

        #region Saved Conversion State

        [Setting]
        [Browsable(false)]
        [DefaultValue(false)]
        public bool SavedToFile { get; set; }

        [Setting]
        [Browsable(false)]
        [DefaultValue(false)]
        public bool FlashHealAdjusted { get; set; }

        #endregion

        [Setting]
        [DefaultValue(100)]
        [Category("生命 %")]
        [DisplayName("% 真言术：盾")]
        [Description("设置成0禁用该法术.")]
        public int PowerWordShield { get; set; }

        [Setting]
        [DefaultValue(95)]
        [Category("生命 %")]
        [DisplayName("% 恢复")]
        [Description("设置成0禁用该法术.")]
        public int Renew { get; set; }

        [Setting]
        [DefaultValue(94)]
        [Category("生命 %")]
        [DisplayName("% 愈合祷言")]
        [Description("设置成0禁用该法术.")]
        public int PrayerOfMending { get; set; }

        [Setting]
        [DefaultValue(0)]
        [Category("生命 %")]
        [DisplayName("% 治疗术")]
        [Description("设置成0禁用该法术.")]
        public int Heal { get; set; }

        [Setting]
        [DefaultValue(40)]
        [Category("生命 %")]
        [DisplayName("% 强效治疗术")]
        [Description("设置成0禁用该法术.")]
        public int GreaterHeal { get; set; }

        [Setting]
        [DefaultValue(20)]
        [Category("生命 %")]
        [DisplayName("% 快速治疗")]
        [Description("设置成0禁用该法术.")]
        public int FlashHeal { get; set; }

        [Setting]
        [DefaultValue(80)]
        [Category("生命 %")]
        [DisplayName("% 联结治疗")]
        [Description("设置成0禁用该法术.")]
        public int BindingHeal { get; set; }

        [Setting]
        [DefaultValue(75)]
        [Category("生命 %")]
        [DisplayName("% 90级天赋")]
        [Description("设置成0禁用该法术.")]
        public int HolyLevel90Talent { get; set; }

        [Setting]
        [DefaultValue(92)]
        [Category("生命 %")]
        [DisplayName("% 圣言术：佑")]
        [Description("设置成0禁用该法术.")]
        public int HolyWordSanctuary { get; set; }

        [Setting]
        [DefaultValue(00)]
        [Category("生命 %")]
        [DisplayName("% 圣言术：静")]
        [Description("设置成0禁用该法术.")]
        public int HolyWordSerenity { get; set; }

        [Setting]
        [DefaultValue(93)]
        [Category("生命 %")]
        [DisplayName("% 治疗之环")]
        [Description("设置成0禁用该法术.")]
        public int CircleOfHealing { get; set; }

        [Setting]
        [DefaultValue(80)]
        [Category("生命 %")]
        [DisplayName("% 治疗祷言")]
        [Description("设置成0禁用该法术.")]
        public int PrayerOfHealing { get; set; }

        [Setting]
        [DefaultValue(70)]
        [Category("生命 %")]
        [DisplayName("% 神圣赞美诗")]
        [Description("设置成0禁用该法术.")]
        public int DivineHymn { get; set; }

        [Setting]
        [DefaultValue(50)]
        [Category("生命 %")]
        [DisplayName("% 守护之魂")]
        [Description("设置成0禁用该法术.")]
        public int GuardianSpirit { get; set; }

        [Setting]
        [DefaultValue(19)]
        [Category("生命 %")]
        [DisplayName("% 虚空转移")]
        [Description("设置成0禁用该法术.")]
        public int VoidShift { get; set; }

        #region Counts

        [Setting]
        [DefaultValue(3)]
        [Category("最低目标数量")]
        [DisplayName("圣言术：佑")]
        [Description("设置需要治疗且在范围内的最低目标数量")]
        public int CountHolyWordSanctuary { get; set; }

        [Setting]
        [DefaultValue(3)]
        [Category("最低目标数量")]
        [DisplayName("90级天赋")]
        [Description("设置需要治疗的最低目标数量")]
        public int CountLevel90Talent { get; set; }

        [Setting]
        [DefaultValue(3)]
        [Category("最低目标数量")]
        [DisplayName("治疗之环")]
        [Description("设置需要治疗的最低目标数量")]
        public int CountCircleOfHealing { get; set; }

        [Setting]
        [DefaultValue(3)]
        [Category("最低目标数量")]
        [DisplayName("治疗祷言")]
        [Description("设置需要治疗的最低目标数量")]
        public int CountPrayerOfHealing { get; set; }

        [Setting]
        [DefaultValue(4)]
        [Category("最低目标数量")]
        [DisplayName("神圣赞美诗")]
        [Description("设置需要治疗的最低目标数量")]
        public int CountDivineHymn { get; set; }

        [Setting]
        [DefaultValue(3)]
        [Category("最低目标数量")]
        [DisplayName("愈合祷言")]
        [Description("设置需要治疗的最低目标数量")]
        public int CountPrayerOfMending { get; set; }

        #endregion
    }

    internal class DiscPriestHealSettings : Singular.Settings.HealerSettings
    {
        private DiscPriestHealSettings()
            : base("", HealingContext.None)
        {
        }

        public DiscPriestHealSettings(HealingContext ctx)
            : base("Priest-Discipline", ctx)
        {

            // we haven't created a settings file yet,
            //  ..  so initialize values for various heal contexts

            if (!SavedToFile)
            {
                if (ctx == Singular.HealingContext.Battlegrounds)
                {
                    Renew = 0;
                    Penance = 80;
                    Heal = 0;
                    PrayerOfMending = 80;
                    GreaterHeal = 0;
                    FlashHeal = 80;
                    BindingHeal = 79;
                    PrayerOfHealing = 90;
                    VoidShift = 35;
                    PainSuppression = 0;
                    DiscLevel90Talent = 85;
                    PowerWordBarrier = 40;
                    CountLevel90Talent = 3;
                    CountPrayerOfHealing = 3;
                    CountPrayerOfMending = 3;
                    CountPowerWordBarrier = 3;
                    AtonementAbovePercent = 90;
                    AtonementAboveCount = 1;
                    AtonementWhenIdle = true;
                }
                else if (ctx == Singular.HealingContext.Instances)
                {
                    Renew = 0;
                    Penance = 80;
                    Heal = 95;
                    PrayerOfMending = 80;
                    GreaterHeal = 60;
                    FlashHeal = 30;
                    BindingHeal = 75;
                    PrayerOfHealing = 90;
                    VoidShift = 29;
                    PainSuppression = 0;
                    DiscLevel90Talent = 85;
                    PowerWordBarrier = 0;
                    CountLevel90Talent = 3;
                    CountPrayerOfHealing = 3;
                    CountPrayerOfMending = 3;
                    CountPowerWordBarrier = 3;
                    AtonementAbovePercent = 90;
                    AtonementAboveCount = 1;
                    AtonementWhenIdle = true;
                }
                else if (ctx == Singular.HealingContext.Raids)
                {
                    Renew = 0;
                    Penance = 80;
                    Heal = 0;
                    PrayerOfMending = 90;
                    GreaterHeal = 50;
                    FlashHeal = 30;
                    BindingHeal = 70;
                    PrayerOfHealing = 89;
                    VoidShift = 20;
                    PainSuppression = 49;
                    DiscLevel90Talent = 89;
                    PowerWordBarrier = 80;
                    CountLevel90Talent = 3;
                    CountPrayerOfHealing = 3;
                    CountPrayerOfMending = 3;
                    CountPowerWordBarrier = 3;
                    AtonementAbovePercent = 90;
                    AtonementAboveCount = 1;
                    AtonementWhenIdle = true;
                }
                // omit case for WoWContext.Normal and let it use DefaultValue() values
            }

            SavedToFile = true;
        }

        [Setting]
        [Browsable(false)]
        [DefaultValue(false)]
        public bool SavedToFile { get; set; }

        [Setting]
        [DefaultValue(100)]
        [Category("治疗")]
        [DisplayName("% 真言术：盾")]
        [Description("低于设置的治疗百分比时使用该技能。设置成0禁用。")]
        public int PowerWordShield { get; set; }

        [Setting]
        [DefaultValue(0)]
        [Category("治疗")]
        [DisplayName("% 恢复")]
        [Description("低于设置的治疗百分比时使用该技能。设置成0禁用。")]
        public int Renew { get; set; }

        [Setting]
        [DefaultValue(0)]
        [Category("治疗")]
        [DisplayName("% 苦修")]
        [Description("低于设置的治疗百分比时使用该技能。设置成0禁用。")]
        public int Penance { get; set; }

        [Setting]
        [DefaultValue(0)]
        [Category("治疗")]
        [DisplayName("% 治疗术")]
        [Description("低于设置的治疗百分比时使用该技能。设置成0禁用。")]
        public int Heal { get; set; }

        [Setting]
        [DefaultValue(90)]
        [Category("治疗")]
        [DisplayName("% 愈合祷言")]
        [Description("低于设置的治疗百分比时使用该技能。设置成0禁用。")]
        public int PrayerOfMending { get; set; }

        [Setting]
        [DefaultValue(60)]
        [Category("治疗")]
        [DisplayName("% 强效治疗术")]
        [Description("低于设置的治疗百分比时使用该技能。设置成0禁用。")]
        public int GreaterHeal { get; set; }

        [Setting]
        [DefaultValue(30)]
        [Category("治疗")]
        [DisplayName("% 快速治疗")]
        [Description("低于设置的治疗百分比时使用该技能。设置成0禁用。")]
        public int FlashHeal { get; set; }

        [Setting]
        [DefaultValue(70)]
        [Category("治疗")]
        [DisplayName("% 联结治疗")]
        [Description("低于设置的治疗百分比时使用该技能。设置成0禁用。")]
        public int BindingHeal { get; set; }

        [Setting]
        [DefaultValue(89)]
        [Category("治疗")]
        [DisplayName("% 治疗祷言")]
        [Description("低于设置的治疗百分比时使用该技能。设置成0禁用。")]
        public int PrayerOfHealing { get; set; }

        [Setting]
        [DefaultValue(20)]
        [Category("治疗")]
        [DisplayName("% 虚空转移")]
        [Description("低于设置的治疗百分比时使用该技能。设置成0禁用。")]
        public int VoidShift { get; set; }

        [Setting]
        [DefaultValue(20)]
        [Category("治疗")]
        [DisplayName("% 痛苦压制")]
        [Description("低于设置的治疗百分比时使用该技能。设置成0禁用。")]
        public int PainSuppression { get; set; }

        [Setting]
        [DefaultValue(89)]
        [Category("生命 %")]
        [DisplayName("% 90级天赋")]
        [Description("低于设置的治疗百分比时使用该技能。设置成0禁用。")]
        public int DiscLevel90Talent { get; set; }

        [Setting]
        [DefaultValue(80)]
        [Category("生命 %")]
        [DisplayName("% 真言术：障")]
        [Description("低于设置的治疗百分比时使用该技能。设置成0禁用。")]
        public int PowerWordBarrier { get; set; }

        [Setting]
        [DefaultValue(75)]
        [Category("生命 %")]
        [DisplayName("% 灵魂护壳")]
        [Description("低于设置的治疗百分比时使用该技能。设置成0禁用。")]
        public int SpiritShell { get; set; }

        [Setting]
        [DefaultValue(3)]
        [Category("最小目标数")]
        [DisplayName("数量，90级天赋")]
        [Description("受到治疗的最小目标数")]
        public int CountLevel90Talent { get; set; }

        [Setting]
        [DefaultValue(3)]
        [Category("最小目标数")]
        [DisplayName("数量，治疗祷言")]
        [Description("受到治疗的最小目标数")]
        public int CountPrayerOfHealing { get; set; }

        [Setting]
        [DefaultValue(3)]
        [Category("最小目标数")]
        [DisplayName("数量，愈合祷言")]
        [Description("受到治疗的最小目标数")]
        public int CountPrayerOfMending { get; set; }

        [Setting]
        [DefaultValue(3)]
        [Category("最小目标数")]
        [DisplayName("数量，真言术：障")]
        [Description("受到治疗的最小目标数")]
        public int CountPowerWordBarrier { get; set; }       

        [Setting]
        [DefaultValue(3)]
        [Category("最小目标数")]
        [DisplayName("数量，灵魂护壳")]
        [Description("Min number of players below health %")]
        public int CountSpiritShell { get; set; }

        [Setting]
        [DefaultValue(90)]
        [Category("戒律")]
        [DisplayName("救赎，仅当高于生命 %")]
        [Description("Only Atonement Healing done above this Health %")]
        public int AtonementAbovePercent { get; set; }

        [Setting]
        [DefaultValue(1)]
        [Category("治疗")]
        [DisplayName("救赎，数量")]
        [Description("Count of Heal Targets below Atonement Only % which allows other heal spells")]
        public int AtonementAboveCount { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("治疗")]
        [DisplayName("Atonement When Idle")]
        [Description("True: DPS with Atonement spells if no healing needed")]
        public bool AtonementWhenIdle { get; set; }


    }
}