﻿
using System.ComponentModel;
using System.IO;
using Singular.ClassSpecific.Rogue;
using Styx.Helpers;

using DefaultValue = Styx.Helpers.DefaultValueAttribute;

namespace Singular.Settings
{
    internal class RogueSettings : Styx.Helpers.Settings
    {
        public RogueSettings()
            : base(Path.Combine(SingularSettings.SingularSettingsPath, "Rogue.xml"))
        {
        }

        [Setting]
        [DefaultValue(false)]
        [Category("一般")]
        [DisplayName("使用疾跑")]
        [Description("疾跑接近目的地或者当不能使用坐骑时")]
        public bool UseSprint { get; set; }

        [Setting]
        [DefaultValue(70)]
        [Category("一般")]
        [DisplayName("恢复，生命 %")]
        [Description("低于设置的生命百分比时使用恢复,设置成 0 则不使用该技能")]
        public int RecuperateHealth { get; set; }

        [Setting]
        [DefaultValue(LethalPoisonType.Auto)]
        [Category("一般")]
        [DisplayName("致命毒药")]
        [Description("致命毒药")]
        public LethalPoisonType LethalPoison { get; set; }

        [Setting]
        [DefaultValue(NonLethalPoisonType.Auto)]
        [Category("一般")]
        [DisplayName("非致命毒药")]
        [Description("非致命毒药")]
        public NonLethalPoisonType NonLethalPoison { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("一般")]
        [DisplayName("总是潜行")]
        [Description("不在战斗时总是潜行。 不会禁用坐骑。 (如果需要你可以在HB设置里进行相关设置)")]
        public bool StealthAlways { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("一般")]
        [DisplayName("吃喝时潜行")]
        [Description("如果正在吃喝则潜行")]
        public bool StealthIfEating { get; set; }

        [Setting]
        [DefaultValue(2)]
        [Category("一般")]
        [DisplayName("刀扇，敌人数量")]
        [Description("当敌人数量超过设置的数量时，使用刀扇来积累连击点数")]
        public int FanOfKnivesCount { get; set; }

        [Setting]
        [DefaultValue(4)]
        [Category("一般")]
        [DisplayName("AOE循环，敌人数量")]
        [Description("当敌人数量超过设置的数量时，使用aoe循环")]
        public int AoeSpellPriorityCount { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("一般")]
        [DisplayName("打断施法")]
        [Description("打断施法")]
        public bool InterruptSpells { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("一般")]
        [DisplayName("使用嫁祸诀窍")]
        [Description("使用嫁祸诀窍")]
        public bool UseTricksOfTheTrade { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("一般")]
        [DisplayName("使用开锁")]
        [Description("需要开启游戏内自动拾取功能; 空闲时自动开启上锁的箱子")]
        public bool UsePickLock { get; set; }

        [Setting]
        [DefaultValue(0)]
        [Category("一般")]
        [DisplayName("闷棍ADD的距离")]
        [Description("闷棍设置范围内可能会add的怪物;设置成0禁用")]
        public int SapAddDistance { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("一般")]
        [DisplayName("闷棍移动的目标")]
        [Description("潜行时使用闷棍来避免跟随移动的目标移动")]
        public bool SapMovingTargetsOnPull { get; set; }


        [Setting]
        [DefaultValue(true)]
        [Category("战斗天赋")]
        [DisplayName("使用割裂作为终结技")]
        [Description("使用割裂作为终结技")]
        public bool CombatUseRuptureFinisher { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("一般")]
        [DisplayName("移动到目标背后")]
        [Description("移动到目标背后开怪或者目标被晕眩或者目标还未选中你")]
        public bool MoveBehindTargets { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("战斗天赋")]
        [DisplayName("使用破甲")]
        [Description("使用破甲")]
        public bool UseExposeArmor { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("一般")]
        [DisplayName("使用加速技能")]
        [Description("当试图脱离战斗时开启加速技能")]
        public bool UseSpeedBuff { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("搜索")]
        [DisplayName("使用搜索")]
        [Description("需要开启游戏内自动拾取功能; 开怪之前先偷")]
        public bool UsePickPocket { get; set; }

        [Setting]
        [DefaultValue(200)]
        [Category("搜索")]
        [DisplayName("搜索之后暂停时间")]
        [Description("搜索之后暂停的时间（ms）(解决搜索技能产生的公公冷却造成自动拾取的卡顿)")]
        public int PostPickPocketPause { get; set; }

        [Setting]
        [DefaultValue(100)]
        [Category("搜索")]
        [DisplayName("搜索之前暂停时间")]
        [Description("搜索之前暂停的时间（ms）(解决系统有延迟)")]
        public int PrePickPocketPause { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("搜索")]
        [DisplayName("允许战斗中使用搜索")]
        [Description("当技能可用且目标可偷时允许战斗中使用搜索")]
        public bool AllowPickPocketInCombat { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("通用")]
        [DisplayName("使用拆卸")]
        [Description("True:冷却结束使用拆卸; False:不使用该技能")]
        public bool UseDimantle { get; set; }
    }
}