﻿
using System.ComponentModel;
using System.IO;
using Styx.Helpers;
using Styx.WoWInternals.WoWObjects;

using DefaultValue = Styx.Helpers.DefaultValueAttribute;
using Styx;
using System.Drawing;
using System;

namespace Singular.Settings
{
    internal class ShamanSettings : Styx.Helpers.Settings
    {
        public ShamanSettings()
            : base(Path.Combine(SingularSettings.SingularSettingsPath, "Shaman.xml"))
        {
        }

        #region Context Late Loading Wrappers

        private ShamanHealSettings _battleground;
        private ShamanHealSettings _instance;
        private ShamanHealSettings _raid;

        [Browsable(false)]
        public ShamanHealSettings Battleground { get { return _battleground ?? (_battleground = new ShamanHealSettings( HealingContext.Battlegrounds )); } }

        [Browsable(false)]
        public ShamanHealSettings Instance { get { return _instance ?? (_instance = new ShamanHealSettings(HealingContext.Instances)); } }

        [Browsable(false)]
        public ShamanHealSettings Raid { get { return _raid ?? (_raid = new ShamanHealSettings(HealingContext.Raids)); } }

        [Browsable(false)]
        public ShamanHealSettings Heal { get { return HealLookup(Singular.SingularRoutine.CurrentWoWContext); } }
        
        public ShamanHealSettings HealLookup( WoWContext ctx)
        {
            if (ctx == WoWContext.Instances)
                return StyxWoW.Me.CurrentMap.IsRaid ? Raid : Instance;

            return Battleground;
        }

        #endregion


        #region Category: Common

        [Setting]
        [DefaultValue(true)]
        [Category("一般")]
        [DisplayName("使用幽魂之狼")]
        [Description("跑步或者室内使用幽魂之狼移动")]
        public bool UseGhostWolf { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("一般")]
        [DisplayName("允许协助治疗")]
        [Description("协助治疗任何生命低于30%的人(团队中自动禁用)")]
        public bool AllowOffHealHeal { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("一般")]
        [DisplayName("使用嗜血/英勇")]
        [Description("适当的时间使用 (禁用移动后会自动禁用该功能)")]
        public bool UseBloodlust { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("一般")]
        [DisplayName("使用武器灌魔")]
        [Description("True: 自动选择并释放合适的武器灌魔, False:防止自动释放武器灌魔")]
        public bool UseWeaponImbues { get; set; }

        [Setting]
        [DefaultValue(15)]
        [Category("一般")]
        [DisplayName("换水盾，魔法 %")]
        [Description("低于设置的魔法百分比时切换水盾")]
        public int TwistWaterShield { get; set; }


        private int _TwistDamageShield;

        [Setting]
        [DefaultValue(25)]
        [Category("一般")]
        [DisplayName("切换 闪电/大地盾，魔法 %")]
        [Description("闪电盾(元素)或者大地之盾(恢复)当魔法高于设定百分比时。提示：至少比水盾高10个百分比。")]
        public int TwistDamageShield 
        { 
            get { return Math.Max( _TwistDamageShield, TwistWaterShield + 10); }
            set { _TwistDamageShield = value; }
        }

        #endregion

        #region Category: Enhancement
        [Setting]
        [DefaultValue(CastOn.All)]
        [Category("增强")]
        [DisplayName("野性狼魂")]
        [Description("选择在何种情形下使用野性狼魂")]
        public CastOn FeralSpiritCastOn  { get; set; }

        [Setting]
        [DefaultValue(75)]
        [Category("增强")]
        [DisplayName("治疗之涌（旋窝）")]
        [Description("低于设置的生命百分比时使用该技能. 设置成100好了就用,设置成0禁用.")]
        public int MaelHealingSurge { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("增强")]
        [DisplayName("禁用旋窝输出")]
        [Description("禁用闪电箭,闪电链以及妖术")]
        public bool AvoidMaelstromDamage { get; set; }

        [Setting]
        [DefaultValue(35)]
        [Category("增强")]
        [DisplayName("PVP时使用旋窝辅助治疗")]
        [Description("低于设置的生命百分比时使用该技能. 设置成100好了就用,设置成0禁用.")]
        public int MaelPvpOffHeal { get; set; }

        #endregion

        #region Category: Self-Healing

        [Setting]
        [DefaultValue(85)]
        [Category("自我治疗")]
        [DisplayName("治疗之泉图腾 %")]
        [Description("低于设置的生命百分比时使用该技能. 设置成100好了就用,设置成0禁用.")]
        public int SelfHealingStreamTotem { get; set; }

        [Setting]
        [DefaultValue(40)]
        [Category("自我治疗")]
        [DisplayName("治疗之涌 %")]
        [Description("低于设置的生命百分比时使用该技能. 设置成100好了就用,设置成0禁用.")]
        public int SelfHealingSurge { get; set; }

        [Setting]
        [DefaultValue(65)]
        [Category("自我治疗")]
        [DisplayName("先祖指引 %")]
        [Description("低于设置的生命百分比时使用该技能. 设置成100好了就用,设置成0禁用.")]
        public int SelfAncestralGuidance { get; set; }

        [Setting]
        [DefaultValue(25)]
        [Category("自我治疗")]
        [DisplayName("先祖迅捷 %")]
        [Description("低于设置的生命百分比时使用该技能. 设置成100好了就用,设置成0禁用.")]
        public int SelfAncestralSwiftnessHeal { get; set; }

        #endregion

        #region Category: Totems

        [Setting]
        [DefaultValue(80)]
        [Category("图腾")]
        [DisplayName("法力之潮图腾，法力 %")]
        [Description("低于设置的魔法百分比时使用法力之潮图腾。设置成 0 则不使用。")]
        public int ManaTideTotemPercent { get; set; }

        #endregion

        #region Talents

        [Setting]
        [DefaultValue(47)]
        [Category("天赋")]
        [DisplayName("星界转移，生命 %")]
        [Description("低于设置的生命百分比时使用星界转移。设置成 0 则不使用。")]
        public int AstralShiftPercent { get; set; }

        [Setting]
        [DefaultValue(47)]
        [Category("天赋")]
        [DisplayName("治疗之潮图腾，生命 %")]
        [Description("低于设置的生命百分比时使用治疗之潮图腾。设置成 0 则不使用。")]
        public int HealingTideTotemPercent { get; set; }

        [Setting]
        [DefaultValue(47)]
        [Category("天赋")]
        [DisplayName("石壁图腾，生命 %")]
        [Description("低于设置的生命百分比时使用石壁图腾。设置成 0 则不使用。")]
        public int StoneBulwarkTotemPercent { get; set; }

        #endregion
    }

    internal class ShamanHealSettings : Singular.Settings.HealerSettings
    {
        private ShamanHealSettings()
            : base("", HealingContext.None)
        {
        }

        public ShamanHealSettings(HealingContext ctx)
            : base("Shaman", ctx)
        {

            // we haven't created a settings file yet,
            //  ..  so initialize values for various heal contexts

            if (!SavedToFile)
            {
                if (ctx == Singular.HealingContext.Battlegrounds)
                {
                    HealingWave = 95;
                    ChainHeal = 90;
                    HealingRain = 93;
                    GreaterHealingWave = 0;
                    Ascendance = 49;
                    SpiritLinkTotem = 50;
                    HealingSurge = 85;
                    AncestralSwiftness = 35;
                    HealingStreamTotem = 90;
                    HealingTideTotem = 55;

                    RollRiptideCount = 0;
                    MinHealingRainCount = 4;
                    MinChainHealCount = 3;
                    MinHealingTideCount = 2;
                }
                else if (ctx == Singular.HealingContext.Instances)
                {
                    HealingWave = 90;
                    ChainHeal = 90;
                    HealingRain = 70;
                    GreaterHealingWave = 70;
                    Ascendance = 48;
                    SpiritLinkTotem = 49;
                    HealingSurge = 60;
                    AncestralSwiftness = 20;
                    HealingStreamTotem = 85;
                    HealingTideTotem = 50;

                    RollRiptideCount = 1;
                    MinHealingRainCount = 4;
                    MinChainHealCount = 3;
                    MinHealingTideCount = 2;
                }
                else if (ctx == Singular.HealingContext.Raids)
                {
                    HealingWave = 93;
                    ChainHeal = 90;
                    HealingRain = 95;
                    GreaterHealingWave = 50;
                    Ascendance = 50;
                    SpiritLinkTotem = 48;
                    HealingSurge = 21;
                    AncestralSwiftness = 20;
                    HealingStreamTotem = 85;
                    HealingTideTotem = 70;

                    RollRiptideCount = 2;
                    MinHealingRainCount = 3;
                    MinChainHealCount = 2;
                    MinHealingTideCount = 4;
                }
                // omit case for WoWContext.Normal and let it use DefaultValue() values
            }

            // adjust Healing Surge if we have not previously 
            if (!HealingSurgeAdjusted && StyxWoW.Me.Level >= 60 && (ctx == HealingContext.Instances || ctx == HealingContext.Raids))
            {
                if ( SavedToFile )
                    Logger.Write(Color.White, "Healing Surge % changed from {0} to {1} for {2}.  Visit Class Config and Save to make permanent.", HealingSurge, AncestralSwiftness + 1, ctx.ToString());

                HealingSurge = AncestralSwiftness + 1;
                HealingSurgeAdjusted = true;
            }

            SavedToFile = true;
        }

        [Setting]
        [Browsable(false)]
        [DefaultValue(false)]
        public bool SavedToFile { get; set; }

        [Setting]
        [Browsable(false)]
        [DefaultValue(false)]
        public bool HealingSurgeAdjusted { get; set; }

        [Setting]
        [DefaultValue(70)]
        [Category("恢复")]
        [DisplayName("% 治疗波")]
        [Description("低于设置的生命百分比时使用该技能.设置成0禁用.")]
        public int HealingWave { get; set; }

        [Setting]
        [DefaultValue(92)]
        [Category("恢复")]
        [DisplayName("% 治疗链")]
        [Description("低于设置的生命百分比时使用该技能.队伍中最少2人，团队中最少3人.设置成0禁用.")]
        public int ChainHeal { get; set; }

        [Setting]
        [DefaultValue(91)]
        [Category("恢复")]
        [DisplayName("% 治疗之雨")]
        [Description("低于设置的生命百分比时使用该技能.队伍中最少3人，团队中最少4人.设置成0禁用.")]
        public int HealingRain { get; set; }

        [Setting]
        [DefaultValue(60)]
        [Category("恢复")]
        [DisplayName("% 强效治疗波")]
        [Description("低于设置的生命百分比时使用该技能.设置成0禁用.")]
        public int GreaterHealingWave { get; set; }

        [Setting]
        [DefaultValue(45)]
        [Category("恢复")]
        [DisplayName("% 升腾")]
        [Description("低于设置的生命百分比时使用该技能.设置成0禁用.")]
        public int Ascendance { get; set; }

        [Setting]
        [DefaultValue(16)]
        [Category("恢复")]
        [DisplayName("% 治疗之涌")]
        [Description("低于设置的生命百分比时使用该技能.设置成0禁用.")]
        public int HealingSurge { get; set; }

        [Setting]
        [DefaultValue(15)]
        [Category("恢复")]
        [DisplayName("% 迅捷治疗!")]
        [Description("低于设置的生命百分比使用迅捷治疗！ (自然迅捷+强效治疗波). 如果设置成0、冷却中、没选择该天赋时禁用.")]
        public int AncestralSwiftness { get; set; }

        [Setting]
        [DefaultValue(48)]
        [Category("恢复")]
        [DisplayName("% 灵魂连接图腾")]
        [Description("低于设置的生命百分比使用该技能. 仅在队伍中有效.设置成0禁用.")]
        public int SpiritLinkTotem { get; set; }

        [Setting]
        [DefaultValue(95)]
        [Category("恢复")]
        [DisplayName("% 治疗之泉图腾")]
        [Description("低于设置的生命百分比时使用该技能.设置成0禁用.")]
        public int HealingStreamTotem { get; set; }

        [Setting]
        [DefaultValue(70)]
        [Category("天赋")]
        [DisplayName("治疗之潮图腾 %")]
        [Description("低于设置的生命百分比时使用该技能.设置成0禁用.")]
        public int HealingTideTotem { get; set; }

        [Setting]
        [DefaultValue(4)]
        [Category("恢复")]
        [DisplayName("维持激流最大目标数")]
        [Description("维持激流的最大目标数 (始终保持坦克身上的激流效果且坦克算在计数中)")]
        public int RollRiptideCount { get; set; }

        [Setting]
        [DefaultValue(4)]
        [Category("恢复")]
        [DisplayName("治疗之雨（数量）")]
        [Description("设置需要治疗且在治疗之雨范围内的最小队友数")]
        public int MinHealingRainCount { get; set; }

        [Setting]
        [DefaultValue(3)]
        [Category("恢复")]
        [DisplayName("治疗链（数量）")]
        [Description("设置需要治疗的最小队友数量")]
        public int MinChainHealCount { get; set; }

        [Setting]
        [DefaultValue(4)]
        [Category("恢复")]
        [DisplayName("治疗之潮图腾（数量）")]
        [Description("设置需要治疗的最小队友数量")]
        public int MinHealingTideCount { get; set; }

        [Setting]
        [DefaultValue(1)]
        [Category("恢复")]
        [DisplayName("灵魂链接图腾（数量）")]
        [Description("设置需要治疗的最小队友数量")]
        public int MinSpiritLinkCount { get; set; }

        [Setting]
        [DefaultValue(3)]
        [Category("恢复")]
        [DisplayName("升腾（数量）")]
        [Description("设置需要治疗的最小队友数量")]
        public int MinAscendanceCount { get; set; }

        [Setting]
        [DefaultValue(90)]
        [Category("恢复")]
        [DisplayName("魔流-闪电链，生命 %")]
        [Description("当团队生命%，可以使用闪电箭回蓝")]
        public int TelluricHealthCast { get; set; }

        [Setting]
        [DefaultValue(80)]
        [Category("恢复")]
        [DisplayName("魔流-闪电链，取消，生命 %")]
        [Description("当团队生命%，不使用闪电箭回蓝")]
        public int TelluricHealthCancel { get; set; }

    }
}