﻿
using System.ComponentModel;
using System.IO;
using System.Linq;
using Styx;
using Styx.Helpers;

using DefaultValue = Styx.Helpers.DefaultValueAttribute;
using Singular.Managers;
using System.Reflection;
using System;
using System.Collections.Generic;
using Singular.Helpers;
using Styx.CommonBot;
using Styx.WoWInternals;

namespace Singular.Settings
{
    enum AllowMovementType
    {
        None,
        ClassSpecificOnly,
        All,
        Auto       
    }

    enum CheckTargets
    {
        None = 0,
        Current,
        All
    }

    enum PurgeAuraFilter
    {
        None = 0,
        Whitelist,
        All
    }

    enum RelativePriority
    {
        None = 0,
        LowPriority,
        HighPriority
    }

    enum TargetingStyle
    {
        None = 0,
        Enable,
        Auto
    }

    enum SelfRessurectStyle
    {
        None = 0,
        Enable,
        Auto
    }

    enum CombatRezTarget
    {
        None = 0,
        All = Tank | Healer | DPS,
        Tank = 1,
        Healer = 2,
        TankOrHealer = Tank | Healer,
        DPS = 4
    }

    internal class SingularSettings : Styx.Helpers.Settings
    {
        private static SingularSettings _instance;

        public SingularSettings()
            : base(Path.Combine(CharacterSettingsPath, "SingularSettings.xml"))
        {
        }

        public static string GlobalSettingsPath
        {
            get
            {
                return Path.Combine(Styx.Common.Utilities.AssemblyDirectory, "Settings");
            }
        }


        public static string CharacterSettingsPath
        {
            get
            {
                string settingsDirectory = Path.Combine(Styx.Common.Utilities.AssemblyDirectory, "Settings");
                return Path.Combine(Path.Combine(settingsDirectory, StyxWoW.Me.RealmName), StyxWoW.Me.Name);
            }
        }


        public static string SingularSettingsPath
        {
            get
            {
                string settingsDirectory = Path.Combine(Styx.Common.Utilities.AssemblyDirectory, "Settings");
                return Path.Combine(Path.Combine(Path.Combine(settingsDirectory, StyxWoW.Me.RealmName), StyxWoW.Me.Name), "Singular");
            }
        }

        public static SingularSettings Instance 
        { 
            get { return _instance ?? (_instance = new SingularSettings()); }
            set { _instance = value; }
        }

        public static bool IsTrinketUsageWanted(TrinketUsage usage)
        {
            return usage == SingularSettings.Instance.Trinket1Usage
                || usage == SingularSettings.Instance.Trinket2Usage
                || usage == SingularSettings.Instance.GloveUsage;
        }

        /// <summary>
        /// Write all Singular Settings in effect to the Log file
        /// </summary>
        public void LogSettings()
        {
            Logger.WriteFile("");

            // reference the internal references so we can display only for our class
            LogSettings("Singular", SingularSettings.Instance);
            LogSettings("HotkeySettings", Hotkeys());
            if (StyxWoW.Me.Class == WoWClass.DeathKnight )  LogSettings("DeathKnightSettings", DeathKnight());
            if (StyxWoW.Me.Class == WoWClass.Druid )        LogSettings("DruidSettings", Druid());
            if (StyxWoW.Me.Class == WoWClass.Hunter )       LogSettings("HunterSettings", Hunter());
            if (StyxWoW.Me.Class == WoWClass.Mage )         LogSettings("MageSettings", Mage());
            if (StyxWoW.Me.Class == WoWClass.Monk )         LogSettings("MonkSettings", Monk());
            if (StyxWoW.Me.Class == WoWClass.Paladin )      LogSettings("PaladinSettings", Paladin());
            if (StyxWoW.Me.Class == WoWClass.Priest )       LogSettings("PriestSettings", Priest());
            if (StyxWoW.Me.Class == WoWClass.Rogue )        LogSettings("RogueSettings", Rogue());
            if (StyxWoW.Me.Class == WoWClass.Shaman)        LogSettings("ShamanSettings", Shaman());
            if (StyxWoW.Me.Class == WoWClass.Warlock)       LogSettings("WarlockSettings", Warlock());
            if (StyxWoW.Me.Class == WoWClass.Warrior)       LogSettings("WarriorSettings", Warrior());

            if (StyxWoW.Me.Specialization == WoWSpec.ShamanRestoration)
            {
                LogSettings("Shaman.Heal.Battleground", Shaman().Battleground);
                LogSettings("Shaman.Heal.Instance", Shaman().Instance);
                LogSettings("Shaman.Heal.Raid", Shaman().Raid);
            }

            if (StyxWoW.Me.Specialization == WoWSpec.PriestHoly)
            {
                LogSettings("Priest.Holy.Heal.Battleground", Priest().HolyBattleground);
                LogSettings("Priest.Holy.Heal.Instance", Priest().HolyInstance);
                LogSettings("Priest.Holy.Heal.Raid", Priest().HolyRaid);
            }

            Logger.WriteFile("====== Evaluated/Dynamic Settings ======");
            Logger.WriteFile("  {0}: {1}", "Debug", SingularSettings.Debug.ToYN());
            Logger.WriteFile("  {0}: {1}", "DisableAllMovement", SingularSettings.Instance.DisableAllMovement.ToYN());
            Logger.WriteFile("  {0}: {1}", "DisableAllTargeting", SingularSettings.DisableAllTargeting.ToYN());
            Logger.WriteFile("  {0}: {1}", "TrivialHealth", Unit.TrivialHealth());
            Logger.WriteFile("");

            if (DisableSpellsWithCooldown == 0)
                Logger.WriteFile("No spells blocked by DisableSpellsWithCooldown");
            else if (!Debug)
                Logger.WriteFile("Spells with cooldown more than {0} secs Blocked by DisableSpellsWithCooldown", DisableSpellsWithCooldown);
            else
            {
                int maxcd = DisableSpellsWithCooldown * 1000;
                Logger.WriteFile("====== Spells Blocked by DisableSpellsWithCooldown  ======");

                using (StyxWoW.Memory.AcquireFrame())
                {
                    foreach (WoWSpell spell in SpellManager.Spells.OrderBy(s => s.Key).Select(s => s.Value))
                    {
                        int cd = Spell.GetBaseCooldown(spell);
                        if (cd >= maxcd)
                        {
                            Logger.WriteFile("  {0} {1}", (cd / 1000).ToString().AlignRight(4), spell.Name);
                        }
                    }
                }
                Logger.WriteFile(" ");
            }
        }

        public void LogSettings(string desc, Styx.Helpers.Settings set)
        {
            if (set == null)
                return;

            Logger.WriteFile("====== {0} Settings ======", desc);
            foreach (var kvp in set.GetSettings())
            {
                Logger.WriteFile("  {0}: {1}", kvp.Key, kvp.Value.ToString());
            }

            Logger.WriteFile("");
        }

        /// <summary>
        /// Obsolete:  Almost all code should reference MovementManager.IsMovementDisabled
        /// .. which will handle Hotkey processing and any context sensitive bot behavior.
        /// .. This setting only retrieves the user setting which typically is insufficient.
        /// </summary>
        [Browsable(false)]
        public bool DisableAllMovement
        {
            get
            {
                if (AllowMovement != AllowMovementType.Auto)
                    return AllowMovement == AllowMovementType.None;

                return SingularRoutine.IsManualMovementBotActive;
            }
        }

        [Browsable(false)]
        public static bool Debug
        {
            get
            {
                return GlobalSettings.Instance.LogLevel >= Styx.Common.LogLevel.Verbose;
            }
        }

        [Browsable(false)]
        public bool DebugSpellCanCast
        {
            get
            {
                return GlobalSettings.Instance.LogLevel >= Styx.Common.LogLevel.Diagnostic;
            }
        }

        [Browsable(false)]
        public static bool Trace
        {
            get
            {
                return Debug && SingularSettings.Instance.EnableDebugTrace;
            }
        }

        [Browsable(false)]
        public static bool DisableAllTargeting
        {
            get
            {
                if (SingularSettings.Instance.TypeOfTargeting != TargetingStyle.Auto)
                    return SingularSettings.Instance.TypeOfTargeting == TargetingStyle.None;

                return SingularRoutine.IsManualMovementBotActive || SingularRoutine.IsDungeonBuddyActive;
            }
        }

        #region Category: Debug
/*
        [Browsable(false)]
        [Setting]
        [DefaultValue(false)]
        [Category("Debug")]
        [DisplayName("Debug Logging")]
        [Description("Enables debug logging from Singular. This will cause quite a bit of spam. Use it for diagnostics only.")]
        public bool EnableDebugLogging { get; set; }
*/
        [Browsable(false)]
        [Setting]
        [DefaultValue(false)]
        [Category("调试")]
        [DisplayName("显示公共冷却")]
        [Description("启用Singular调试日志中的公共冷却信息.调试日志功能也需要同时开启。")]
        public bool EnableDebugLoggingGCD { get; set; }

        [Browsable(false)]
        [Setting]
        [DefaultValue(false)]
        [Category("调试")]
        [DisplayName("调试跟踪")]
        [Description("EXTREMELY VERBOSE!! Enables logging of entry/exit into each behavior. Only use if instructed or you prefer slower response times!")]
        public bool EnableDebugTrace { get; set; }

        #endregion

        #region Window Layout

        [Browsable(false)]
        [Setting]
        [DefaultValue(337)]
        public int FormHeight { get; set; }

        [Browsable(false)]
        [Setting]
        [DefaultValue(378)]
        public int FormWidth { get; set; }

        [Browsable(false)]
        [Setting]
        [DefaultValue(0)]
        public int FormTabIndex { get; set; }


        #endregion

        #region Category: Movement

        [Setting]
        [DefaultValue(AllowMovementType.Auto)]
        [Category("移动")]
        [DisplayName("允许移动")]
        [Description("允许CC控制角色移动. None: 防止所有移动; ClassSpecificOnly: 仅使用对冲/英勇跳跃/闪现/逃脱/等; All: 允许所有移动; Auto: 当<懒人模式>启用时和<None>一样,其他情况下同<All>。")]
        public AllowMovementType AllowMovement { get; set; }

        [Setting]
        [DefaultValue(12)]
        [Category("移动")]
        [DisplayName("近战下坐骑距离")]
        [Description("当与目标的距离小于设置的距离时，近战下坐骑")]
        public int MeleeDismountRange { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("移动")]
        [DisplayName("使用移动施法技能")]
        [Description("True: 在移动中使用非瞬发法术前先尝试使用灵魂行者之赐,浮冰,基尔加丹的狡诈,等。")]
        public bool UseCastWhileMovingBuffs { get; set; }

        #endregion 

        #region Category: Consumables

        [Setting]
        [DefaultValue(65)]
        [Category("消耗品")]
        [DisplayName("使用食物，生命 %")]
        [Description("设置使用食物的最低生命百分比。")]
        public int MinHealth { get; set; }

        [Setting]
        [DefaultValue(65)]
        [Category("消耗品")]
        [DisplayName("使用饮料，魔法 %")]
        [Description("设置使用饮料的最低魔法百分比。")]
        public int MinMana { get; set; }

        [Setting]
        [DefaultValue(30)]
        [Category("消耗品")]
        [DisplayName("使用药水，生命 %")]
        [Description("设置使用治疗药水/饰品/治疗石的最低生命百分比。")]
        public int PotionHealth { get; set; }

        [Setting]
        [DefaultValue(30)]
        [Category("消耗品")]
        [DisplayName("使用药水，魔法 %")]
        [Description("设置使用魔法药水/饰品的最低魔法百分比。(对所有形式的能量资源均有效)")]
        public int PotionMana { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("消耗品")]
        [DisplayName("使用绷带")]
        [Description("使用背包中的绷带治疗。")]
        public bool UseBandages { get; set; }

        #endregion

        #region Category: Avoidance

        [Setting]
        [DefaultValue(true)]
        [Category("规避")]
        [DisplayName("允许逃脱")]
        [Description("允许使用逃脱 (or equiv)")]
        public bool DisengageAllowed { get; set; }

        [Setting]
        [DefaultValue(70)]
        [Category("规避")]
        [DisplayName("逃脱，生命 %")]
        [Description("逃脱(or equiv)如果生命低于设置的百分比并且目标进入近战范围")]
        public int DisengageHealth { get; set; }

        [Setting]
        [DefaultValue(2)]
        [Category("规避")]
        [DisplayName("逃脱，敌人数量")]
        [Description("逃脱(or equiv)如果超过设置的敌人的数量进入近战范围")]
        public int DisengageMobCount { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("规避")]
        [DisplayName("允许风筝")]
        [Description("允许风筝目标")]
        public bool KiteAllow { get; set; }

        [Setting]
        [DefaultValue(50)]
        [Category("规避")]
        [DisplayName("风筝，生命 %")]
        [Description("风筝如果生命低于设置的百分比并且目标进入近战范围")]
        public int KiteHealth { get; set; }

        [Setting]
        [DefaultValue(2)]
        [Category("规避")]
        [DisplayName("风筝，敌人数量")]
        [Description("风筝如果超过设置的敌人的数量进入近战范围")]
        public int KiteMobCount { get; set; }

        [Setting]
        [DefaultValue(8)]
        [Category("规避")]
        [DisplayName("规避距离")]
        [Description("Only mobs within this distance that are attacking you count towards Disengage/Kite mob counts")]
        public int AvoidDistance { get; set; }

        [Browsable(false)]
        [Setting]
        [DefaultValue(false)]
        [Category("规避")]
        [DisplayName("风筝，跳跃转身")]
        [Description("执行跳转攻击来风筝 (目前仅支持猎人)")]
        public bool JumpTurnAllow { get; set; }

        #endregion

        #region Category: General

        [Setting]
        [DefaultValue(false)]
        [Category("通用")]
        [DisplayName("使用帧锁定")]
        [Description("强制使用帧锁定。主要用于某些不支持帧锁定的软件模式")]
        public bool UseFrameLock { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("通用")]
        [DisplayName("等待复活虚弱")]
        [Description("等待复活虚弱状态结束")]
        public bool WaitForResSickness { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("通用")]
        [DisplayName("禁用战斗外行为")]
        [Description("允许它会禁用战斗外行为。例如 (吃喝恢复，加职业增益)")]
        public bool DisableNonCombatBehaviors { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("通用")]
        [DisplayName("禁用当在任务运输工具时")]
        [Description("True:CC忽略任务模式的一切指令当人物出于任务特殊的交通工具里时; False: CC会尝试战斗/治疗，当任务模式让它这么做时。")]
        public bool DisableInQuestVehicle { get; set; }

        [Setting]
        [DefaultValue(0)]
        [Category("通用")]
        [DisplayName("法术冷却时禁用(秒)")]
        [Description("Do not cast any spell with this cooldown or greater; 0 will cast all spells regardless of cooldown")]
        public int DisableSpellsWithCooldown { get; set; }

        [Setting]
        [DefaultValue(SelfRessurectStyle.Auto )]
        [Category("通用")]
        [DisplayName("自动复活")]
        [Description("Auto: 自动复活(重生/灵魂石)除非禁用了移动功能, Enable:永远自动复活当可用时, None: 从不自动复活")]
        public SelfRessurectStyle SelfRessurect { get; set; }

        #endregion

        #region Category: Pets

        [Setting]
        [DefaultValue(false)]
        [Category("宠物")]
        [DisplayName("禁用宠物")]
        [Description("允许该项会禁用宠物")]
        public bool DisablePetUsage { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("宠物")]
        [DisplayName("单人: 宠物吸引adds")]
        [Description("True: 单人模式下，宠物会主动嘲讽正在攻击玩家的怪物")]
        public bool PetTankAdds { get; set; }

        #endregion 

        #region Category: Group Healing / Support

        [Setting]
        [DefaultValue(95)]
        [Category("队伍治疗/支援")]
        [DisplayName("忽略的治疗目标")]
        [Description("设置一个生命值百分比，如果治疗目标的生命高于该值即使正在治疗也会打断读条停止治疗.")]
        public int IgnoreHealTargetsAboveHealth { get; set; }

        [Setting]
        [DefaultValue(75)]
        [Category("队伍治疗/支援")]
        [DisplayName("治疗目标最大距离")]
        [Description("设置寻找治疗目标的最大距离(最远为: 100)")]
        public int MaxHealTargetRange { get; set; }

        [Setting]
        [DefaultValue(25)]
        [Category("队伍治疗/支援")]
        [DisplayName("保持在坦克附近")]
        [Description("在我们向坦克移动前的最大距离 (最大为: 100)")]
        public int StayNearTankRange { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("队伍治疗/支援")]
        [DisplayName("保持在坦克附近")]
        [Description("如果没人需要治疗则始终保持在能治疗到坦克的范围内移动")]
        public bool StayNearTank { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("队伍治疗/支援")]
        [DisplayName("治疗宠物")]
        [Description("True: 会对宠物进行治疗 (禁用该项不会影响主人治疗自己的宠物)")]
        public bool IncludePetsAsHealTargets { get; set; }

        [Setting]
        [DefaultValue(RelativePriority.LowPriority )]
        [Category("队伍治疗/支援")]
        [DisplayName("驱散")]
        [Description("驱散有害状态")]
        public RelativePriority DispelDebuffs { get; set; }

        [Setting]
        [DefaultValue(CombatRezTarget.TankOrHealer)]
        [Category("队伍治疗/支援")]
        [DisplayName("战斗复活目标")]
        [Description("None: 禁用战复; 其他设置是否战复取决于他的团队职责")]
        public CombatRezTarget CombatRezTarget { get; set; }

        [Setting]
        [DefaultValue(2)]
        [Category("队伍治疗/支援")]
        [DisplayName("战斗复活延迟")]
        [Description("等待(秒)再对队友施放战斗复活法术")]
        public int CombatRezDelay { get; set; }

        #endregion

        #region Category: Healing

        #endregion

        #region Category: Items

        [Setting]
        [DefaultValue(true)]
        [Category("物品")]
        [DisplayName("使用炼金合剂")]
        [Description("使用炼金合剂")]
        public bool UseAlchemyFlasks { get; set; }

        [Setting]
        [DefaultValue(TrinketUsage.Never)]
        [Category("物品")]
        [DisplayName("使用饰品 1")]
        public TrinketUsage Trinket1Usage { get; set; }

        [Setting]
        [DefaultValue(TrinketUsage.Never)]
        [Category("物品")]
        [DisplayName("使用饰品 2")]
        public TrinketUsage Trinket2Usage { get; set; }

        [Setting]
        [DefaultValue(TrinketUsage.OnCooldownInCombat)]
        [Category("物品")]
        [DisplayName("使用物品附魔")]
        public TrinketUsage GloveUsage { get; set; }

        #endregion

        #region Category: Racials

        [Setting]
        [DefaultValue(true)]
        [Category("种族技能")]
        [DisplayName("使用种族技能")]
        public bool UseRacials { get; set; }

        [Setting]
        [DefaultValue(30)]
        [Category("种族技能")]
        [DisplayName("使用纳鲁的祝福，生命 %")]
        [Description("当生命低于设置的生命百分比时使用纳鲁的祝福。")]
        public int GiftNaaruHP { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("种族技能")]
        [DisplayName("隐遁清仇恨")]
        [Description("当在队伍中(且不是坦克),使用隐遁清仇恨。")]
        public bool ShadowmeldThreatDrop { get; set; }
        
        #endregion

        #region Category: Tanking

        [Setting]
        [DefaultValue(false)]
        [Category("坦克")]
        [DisplayName("禁用目标切换")]
        [Description("True:坦克时保持当前目标.  False: 切换目标保持在队友中的仇恨最高")]
        public bool DisableTankTargetSwitching { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("坦克")]
        [DisplayName("允许嘲讽技能")]
        public bool EnableTaunting { get; set; }

        #endregion

        #region Category: Targeting

        [Setting]
        [DefaultValue(TargetingStyle.Auto)]
        [Category("选怪")]
        [DisplayName("使用Singular选怪")]
        [Description("None: 禁用, Enable: 智能切换; Auto: 地下城模式和其他辅助类模式下禁用,其他情况下启用。")]
        public TargetingStyle TypeOfTargeting { get; set; }

        #endregion

        #region Category: Enemy Control

        [Setting]
        [DefaultValue(CheckTargets.Current)]
        [Category("敌人控制")]
        [DisplayName("净化目标")]
        [Description("None: 禁用, Current: 仅作用于目标, All: 范围内正面的所有目标")]
        public CheckTargets PurgeTargets { get; set; }

        [Setting]
        [DefaultValue(PurgeAuraFilter.Whitelist)]
        [Category("敌人控制")]
        [DisplayName("净化增益状态")]
        [Description("False: 禁用, Current: 仅作用于目标, Other: 范围内正面的所有目标")]
        public PurgeAuraFilter PurgeBuffs { get; set; }

        [Setting]
        [DefaultValue(CheckTargets.All)]
        [Category("敌人控制")]
        [DisplayName("打断目标")]
        [Description("None: 禁用, Current: 仅作用于目标, All: 范围内正面的所有目标")]
        public CheckTargets InterruptTarget { get; set; }

        [Setting]
        [DefaultValue(50)]
        [Category("敌人控制")]
        [DisplayName("忽略怪物最大生命百分比")]
        [Description("怪物的生命百分比比你少被认为是微不足道的")]
        public int TrivialMaxHealthPcnt { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("敌人控制")]
        [DisplayName("使用aoe技能")]
        [Description("True: 需要时使用多目标伤害技能; False: 仅对当前目标使用单体技能")]
        public bool AllowAOE { get; set; }

        #endregion

        #region Class Late-Loading Wrappers

        // Do not change anything within this region.
        // It's written so we ONLY load the settings we're going to use.
        // There's no reason to load the settings for every class, if we're only executing code for a Druid.

        private DeathKnightSettings _dkSettings;

        private DruidSettings _druidSettings;

        private HunterSettings _hunterSettings;

        private MageSettings _mageSettings;
		
		private MonkSettings _monkSettings;
		
        private PaladinSettings _pallySettings;

        private PriestSettings _priestSettings;

        private RogueSettings _rogueSettings;

        private ShamanSettings _shamanSettings;

        private WarlockSettings _warlockSettings;

        private WarriorSettings _warriorSettings;

        private HotkeySettings _hotkeySettings;

        // late-binding interfaces 
        // -- changed from readonly properties to methods as GetProperties() in SaveToXML() was causing all classes configs to load
        // -- this was causing Save to write a DeathKnight.xml file for all non-DKs for example
        internal DeathKnightSettings DeathKnight() { return _dkSettings ?? (_dkSettings = new DeathKnightSettings()); } 
        internal DruidSettings Druid() { return _druidSettings ?? (_druidSettings = new DruidSettings()); }
        internal HunterSettings Hunter() { return _hunterSettings ?? (_hunterSettings = new HunterSettings()); }
        internal MageSettings Mage() { return _mageSettings ?? (_mageSettings = new MageSettings()); }
        internal MonkSettings Monk() { return _monkSettings ?? (_monkSettings = new MonkSettings()); }
        internal PaladinSettings Paladin() { return _pallySettings ?? (_pallySettings = new PaladinSettings()); }
        internal PriestSettings Priest() { return _priestSettings ?? (_priestSettings = new PriestSettings()); }
        internal RogueSettings Rogue() { return _rogueSettings ?? (_rogueSettings = new RogueSettings()); }
        internal ShamanSettings Shaman() { return _shamanSettings ?? (_shamanSettings = new ShamanSettings()); }
        internal WarlockSettings Warlock() { return _warlockSettings ?? (_warlockSettings = new WarlockSettings()); }
        internal WarriorSettings Warrior() { return _warriorSettings ?? (_warriorSettings = new WarriorSettings()); }
        internal HotkeySettings Hotkeys() { return _hotkeySettings ?? (_hotkeySettings = new HotkeySettings()); }

        #endregion
    }

}