﻿
using System;
using System.ComponentModel;
using System.IO;
using Styx.Helpers;

using DefaultValue = Styx.Helpers.DefaultValueAttribute;
using Styx;
using Styx.CommonBot;
using Singular.Managers;


namespace Singular.Settings
{
    public enum WarlockPet
    {
        None        = 0,        
        Auto        = 1,
        Imp         = 23,       // Pet.CreatureFamily.Id
        Voidwalker  = 16,
        Succubus    = 17,
        Felhunter   = 15,
        Felguard    = 29,
        Other       = 99999     // a quest or other pet forced upon us for some reason
    }

    public enum Soulstone
    {
        None = 0,
        Auto,
        Self,
        Ressurect
    }

    internal class WarlockSettings : Styx.Helpers.Settings
    {

        public WarlockSettings()
            : base(Path.Combine(SingularSettings.SingularSettingsPath, "Warlock.xml"))
        {
        }

        [Setting]
        [DefaultValue(WarlockPet.Auto)]
        [Category("一般")]
        [DisplayName("宠物")]
        [Description("选择召唤何种宠物. Auto：自动选择。如果设置的不可用那么会召唤虚空行者")]
        public WarlockPet Pet { get; set; }

        [Setting]
        [DefaultValue(40)]
        [Category("宠物")]
        [DisplayName("生命通道，宠物生命 %")]
        [Description("战斗中宠物低于设置的生命百分比时使用生命通道")]
        public int HealthFunnelCast { get; set; }

        [Setting]
        [DefaultValue(95)]
        [Category("宠物")]
        [DisplayName("取消生命通道，宠物生命 %")]
        [Description("战斗中宠物生命高于设置的生命百分比时取消正在使用的生命通道")]
        public int HealthFunnelCancel { get; set; }

        [Setting]
        [DefaultValue(75)]
        [Category("宠物")]
        [DisplayName("生命通道，宠物生命 %（战斗外）")]
        [Description("战斗外宠物低于设置的生命百分比时使用生命通道")]
        public int HealthFunnelRest { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("宠物")]
        [DisplayName("使用缴械")]
        [Description("True:一冷却就使用缴械; False:不使用")]
        public bool UseDisarm { get; set; }

        [Setting]
        [DefaultValue(20)]
        [Category("一般")]
        [DisplayName("生命吸取，生命 %")]
        [Description("低于设置生命百分比时使用生命吸取")]
        public int DrainLifePercentage { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("一般")]
        [DisplayName("使用恐惧")]
        [Description("当生命过低或者控制add时使用恐惧")]
        public bool UseFear { get; set; }

        [Setting]
        [DefaultValue(3)]
        [Category("一般")]
        [DisplayName("使用恐惧，数量")]
        [Description("当攻击术士的敌人达到设置数量时(不含宠物);设置成0禁用数量检查")]
        public int UseFearCount { get; set; }

        [Setting]
        [DefaultValue(Soulstone.Auto)]
        [Category("一般")]
        [DisplayName("使用灵魂石")]
        [Description("Auto: 在队伍中=复活, 普通状态/战场中=自己, 禁止移动=不使用")]
        public Soulstone UseSoulstone { get; set; }

        [Setting]
        [DefaultValue(750)]
        [Category("恶魔")]
        [DisplayName("切换至施法者形态的恶魔之怒")]
        [Description("切换至施法者形态当恶魔之怒低于(0 - 1000)")]
        public int FurySwitchToCaster { get; set; }

        [Setting]
        [DefaultValue(900)]
        [Category("恶魔")]
        [DisplayName("切换至恶魔形态的恶魔之怒")]
        [Description("切换至恶魔形态当恶魔之怒高于(0 - 1000)")]
        public int FurySwitchToDemon { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("恶魔")]
        [DisplayName("使用恶魔之跃")]
        [Description("使用恶魔之跃来摆脱近战")]
        public bool UseDemonicLeap { get; set; }

        public enum SpellPriority
        {
            Noxxic = 1,
            IcyVeins = 2
        }

        [Setting]
        [DefaultValue(SpellPriority.Noxxic )]
        [Category("毁灭")]
        [DisplayName("法术优先选择")]
        public SpellPriority DestructionSpellPriority { get; set; }



#region Setting Helpers


#endregion

    }
}