﻿
using System.ComponentModel;
using System.IO;
using Styx.Helpers;

using DefaultValue = Styx.Helpers.DefaultValueAttribute;

namespace Singular.Settings
{
    public enum WarriorStance
    {
        Auto,
        BattleStance        = Styx.ShapeshiftForm.BattleStance ,
        BerserkerStance     = Styx.ShapeshiftForm.BerserkerStance ,
        DefensiveStance     = Styx.ShapeshiftForm.DefensiveStance 
    }

    public enum WarriorShout
    {
        CommandingShout,
        BattleShout
    }

    internal class WarriorSettings : Styx.Helpers.Settings
    {
        public WarriorSettings()
            : base(Path.Combine(SingularSettings.SingularSettingsPath, "Warrior.xml"))
        {
        }

        #region Protection

        public enum SpellPriority
        {
            Noxxic = 1,
            IcyVeins = 2,
            ElitistJerks = 3
        }

        [Setting]
        [DefaultValue(SpellPriority.Noxxic)]
        [Category("武器")]
        [DisplayName("法术优先选择")]
        public SpellPriority ArmsSpellPriority { get; set; }

        [Setting]
        [DefaultValue(50)]
        [Category("防御")]
        [DisplayName("狂怒回复，生命 %")]
        [Description("低于设置的生命百分比时使用狂怒回复")]
        public int WarriorEnragedRegenerationHealth { get; set; }

        [Setting]
        [DefaultValue(20)]
        [Category("防御")]
        [DisplayName("盾墙，生命 %")]
        [Description("低于设置的生命百分比时使用盾墙")]
        public int WarriorShieldWallHealth { get; set; }


        [Setting]
        [DefaultValue(40)]
        [Category("防御")]
        [DisplayName("破釜沉舟，生命 %")]
        [Description("低于设置的生命百分比时使用破釜沉舟")]
        public int WarriorLastStandHealth  { get; set; }

        [Setting]
        [DefaultValue(50)]
        [Category("防御")]
        [DisplayName("盾牌格挡，生命 %")]
        [Description("低于设置的生命百分比时使用盾牌格挡")]
        public int WarriorShieldBlockHealth { get; set; }

        [Setting]
        [DefaultValue(80)]
        [Category("防御")]
        [DisplayName("盾牌屏障，生命 %")]
        [Description("低于设置的生命百分比时使用盾牌屏障")]
        public int WarriorShieldBarrierHealth { get; set; }

        #endregion

        #region DPS

        [Setting]
        [DefaultValue(true)]
        [Category("输出")]
        [DisplayName("使用打断")]
        [Description("True / False：如果你希望CC帮你打断施法的话")]
        public bool UseWarriorInterrupts { get; set; }

        [Setting]
        [DefaultValue(true)]
        [Category("输出")]
        [DisplayName("缓速")]
        [Description("True / False：如果你希望cc使用减速技能的话。如： 断筋,刺耳怒吼")]
        public bool UseWarriorSlows { get; set; }
        
        [Setting]
        [DefaultValue(true)]
        [Category("输出")]
        [DisplayName("使用冲锋/英勇跳跃?")]
        [Description("True / False：如果你希望cc使用这些迅速接近技能")]
        public bool UseWarriorCloser { get; set; }
        #endregion

        [Setting]
        [DefaultValue(WarriorStance.Auto)]
        [Category("输出")]
        [DisplayName("战士输出姿态")]
        [Description("输出时使用的姿态。预料到有伤害来临时会使用战斗姿态否则一直使用狂暴姿态。防御战士会一直使用防御姿态。")]
        public WarriorStance Stance { get; set; }

        [Setting]
        [DefaultValue(WarriorShout.BattleShout)]
        [Category("一般")]
        [DisplayName("战士怒吼")]
        [Description("怒吼被用来保持身上的增益状态,或者在低怒气的时候使用")]
        public WarriorShout Shout { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("一般")]
        [DisplayName("乘胜追击")]
        [Description("True: 技能一冷却就使用乘胜追击/胜利在望，无论当时的生命百分比是多少。")]
        public bool VictoryRushOnCooldown { get; set; }

        [Setting]
        [DefaultValue(false)]
        [Category("一般")]
        [DisplayName("缴械")]
        [Description("True: 技能一冷却就使用缴械; False: 不使用该技能")]
        public bool UseDisarm { get; set; }

    }
}